# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from numbers import Number

import pytest
import numpy as np
from hypothesis import assume, settings
from hypothesis import strategies as hyst

from pymor.algorithms.basic import almost_equal
from pymor.vectorarrays.interface import VectorSpace
from pymor.tools import floatcmp
from pymor.tools.floatcmp import float_cmp
from pymortests.pickling import assert_picklable_without_dumps_function
import pymortests.strategies as pyst


def ind_complement(v, ind):
    if isinstance(ind, Number):
        ind = [ind]
    elif type(ind) is slice:
        ind = range(*ind.indices(len(v)))
    l = len(v)
    return sorted(set(range(l)) - {i if i >= 0 else l+i for i in ind})


def indexed(v, ind):
    if ind is None:
        return v
    elif type(ind) is slice:
        return v[ind]
    elif isinstance(ind, Number):
        return v[[ind]]
    elif len(ind) == 0:
        return np.empty((0, v.shape[1]), dtype=v.dtype)
    else:
        return v[ind]


def ind_to_list(v, ind):
    if type(ind) is slice:
        return list(range(*ind.indices(len(v))))
    elif not hasattr(ind, '__len__'):
        return [ind]
    else:
        return ind


@pyst.given_vector_arrays()
def test_empty(vector_array):
    with pytest.raises(Exception):
        vector_array.empty(-1)
    for r in (0, 1, 100):
        v = vector_array.empty(reserve=r)
        assert v.space == vector_array.space
        assert len(v) == 0
        try:
            assert v.to_numpy().shape == (0, v.dim)
        except NotImplementedError:
            pass


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices)
def test_print(vectors_and_indices):
    v, ind = vectors_and_indices
    assert len(str(v))
    assert len(repr(v))
    assert len(str(v[ind]))
    assert len(repr(v[ind]))


@pyst.given_vector_arrays()
def test_zeros(vector_array):
    with pytest.raises(Exception):
        vector_array.zeros(-1)
    for c in (0, 1, 2, 30):
        v = vector_array.zeros(count=c)
        assert v.space == vector_array.space
        assert len(v) == c
        if min(v.dim, c) > 0:
            assert max(v.sup_norm()) == 0
            assert max(v.norm()) == 0
        try:
            assert v.to_numpy().shape == (c, v.dim)
            assert np.allclose(v.to_numpy(), np.zeros((c, v.dim)))
        except NotImplementedError:
            pass


@pyst.given_vector_arrays()
def test_ones(vector_array):
    with pytest.raises(Exception):
        vector_array.ones(-1)
    for c in (0, 1, 2, 30):
        v = vector_array.ones(count=c)
        assert v.space == vector_array.space
        assert len(v) == c
        if min(v.dim, c) > 0:
            assert np.allclose(v.sup_norm(), np.ones(c))
            assert np.allclose(v.norm(), np.full(c, np.sqrt(v.dim)))
        try:
            assert v.to_numpy().shape == (c, v.dim)
            assert np.allclose(v.to_numpy(), np.ones((c, v.dim)))
        except NotImplementedError:
            pass


@pyst.given_vector_arrays()
def test_full(vector_array):
    with pytest.raises(Exception):
        vector_array.full(9, -1)
    for c in (0, 1, 2, 30):
        for val in (-1e-3, 0, 7):
            v = vector_array.full(val, count=c)
            assert v.space == vector_array.space
            assert len(v) == c
            if min(v.dim, c) > 0:
                assert np.allclose(v.sup_norm(), np.full(c, abs(val)))
                assert np.allclose(v.norm(), np.full(c, np.sqrt(val**2 * v.dim)))
            try:
                assert v.to_numpy().shape == (c, v.dim)
                assert np.allclose(v.to_numpy(), np.full((c, v.dim), val))
            except NotImplementedError:
                pass


@pyst.given_vector_arrays(realizations=hyst.integers(min_value=0, max_value=30),
                          low=hyst.floats(allow_infinity=False, allow_nan=False),
                          high=hyst.floats(allow_infinity=False, allow_nan=False))
def test_random_uniform(vector_array, realizations, low, high):
    # avoid Overflow in np.random.RandomState.uniform
    assume(np.isfinite(high-low))
    with pytest.raises(Exception):
        vector_array.random(-1)
    c = realizations
    if c > 0 and high <= low:
        with pytest.raises(ValueError):
            vector_array.random(c, low=low, high=high)
        return
    seed = 123
    try:
        v = vector_array.random(c, low=low, high=high, seed=seed)
    except ValueError as e:
        if high <= low:
            return
        raise e
    assert v.space == vector_array.space
    assert len(v) == c
    if min(v.dim, c) > 0:
        assert np.all(v.sup_norm() < max(abs(low), abs(high)))
    try:
        x = v.to_numpy()
        assert x.shape == (c, v.dim)
        assert np.all(x < high)
        assert np.all(x >= low)
    except NotImplementedError:
        pass
    vv = vector_array.random(c, distribution='uniform', low=low, high=high, seed=seed)
    assert np.allclose((v - vv).sup_norm(), 0.)


@pyst.given_vector_arrays(realizations=hyst.integers(min_value=0, max_value=30),
                          loc=hyst.floats(allow_infinity=False, allow_nan=False),
                          scale=hyst.floats(allow_infinity=False, allow_nan=False))
def test_random_normal(vector_array, realizations, loc, scale):
    with pytest.raises(Exception):
        vector_array.random(-1)
    c = realizations
    if c > 0 > scale:
        with pytest.raises(ValueError):
            vector_array.random(c, 'normal', loc=loc, scale=scale)
        return
    seed = 123
    try:
        v = vector_array.random(c, 'normal', loc=loc, scale=scale, seed=seed)
    except ValueError as e:
        if scale <= 0:
            return
        raise e
    assert v.space == vector_array.space
    assert len(v) == c
    try:
        x = v.to_numpy()
        assert x.shape == (c, v.dim)
        import scipy.stats
        n = x.size
        if n == 0:
            return
        # test for expected value
        norm = scipy.stats.norm()
        gamma = 1 - 1e-7
        alpha = 1 - gamma
        lower = np.sum(x)/n - norm.ppf(1 - alpha/2) * scale / np.sqrt(n)
        upper = np.sum(x)/n + norm.ppf(1 - alpha/2) * scale / np.sqrt(n)
        floatcmp.bounded(lower, upper, loc)
    except NotImplementedError:
        pass
    vv = vector_array.random(c, 'normal', loc=loc, scale=scale, seed=seed)
    data = vv.to_numpy()
    # due to scaling data might actually now include nan or inf
    assume(not np.isnan(data).any())
    assume(not np.isinf(data).any())
    assert np.allclose((v - vv).sup_norm(), 0.)


@pyst.given_vector_arrays()
def test_from_numpy(vector_array):
    try:
        d = vector_array.to_numpy()
    except NotImplementedError:
        return
    try:
        v = vector_array.space.from_numpy(d)
        assert np.allclose(d, v.to_numpy())
    except NotImplementedError:
        pass


@pyst.given_vector_arrays()
def test_shape(vector_array):
    assert len(vector_array) >= 0
    assert vector_array.dim >= 0
    try:
        assert vector_array.to_numpy().shape == (len(vector_array), vector_array.dim)
    except NotImplementedError:
        pass


@pyst.given_vector_arrays()
def test_space(vector_array):
    assert isinstance(vector_array.space, VectorSpace)
    assert vector_array in vector_array.space


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices)
def test_getitem_repeated(vectors_and_indices):
    v, ind = vectors_and_indices
    v_ind = v[ind]
    v_ind_copy = v_ind.copy()
    assert not v_ind_copy.is_view
    for ind_ind in pyst.valid_inds(v_ind, random_module=False):
        v_ind_ind = v_ind[ind_ind]
        assert np.all(almost_equal(v_ind_ind, v_ind_copy[ind_ind]))


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices)
def test_copy(vectors_and_indices):
    v, ind = vectors_and_indices
    for deep in (True, False):
        if ind is None:
            c = v.copy(deep)
            assert len(c) == len(v)
        else:
            c = v[ind].copy(deep)
            assert len(c) == v.len_ind(ind)
        assert c.space == v.space
        if ind is None:
            assert np.all(almost_equal(c, v))
        else:
            assert np.all(almost_equal(c, v[ind]))
        try:
            assert np.allclose(c.to_numpy(), indexed(v.to_numpy(), ind))
        except NotImplementedError:
            pass


@pyst.given_vector_arrays()
def test_copy_repeated_index(vector_array):
    v = vector_array
    if len(v) == 0:
        return
    ind = [int(len(v) * 3 / 4)] * 2
    for deep in (True, False):
        c = v[ind].copy(deep)
        assert almost_equal(c[0], v[ind[0]])
        assert almost_equal(c[1], v[ind[0]])
        try:
            assert indexed(v.to_numpy(), ind).shape == c.to_numpy().shape
        except NotImplementedError:
            pass
        c[0].scal(2.)
        assume(c[0].norm() != np.inf)
        assert almost_equal(c[1], v[ind[0]])
        assert float_cmp(c[0].norm(), 2 * v[ind[0]].norm())
        try:
            assert indexed(v.to_numpy(), ind).shape == c.to_numpy().shape
        except NotImplementedError:
            pass


@pyst.given_vector_arrays(count=2, index_strategy=pyst.pairs_both_lengths)
def test_append(vectors_and_indices):
    (v1, v2), (_, ind) = vectors_and_indices
    len_v1 = len(v1)
    c1, c2 = v1.copy(), v2.copy()
    c1.append(c2[ind])
    len_ind = v2.len_ind(ind)
    ind_complement_ = ind_complement(v2, ind)
    assert len(c1) == len_v1 + len_ind
    assert np.all(almost_equal(c1[len_v1:len(c1)], c2[ind]))
    try:
        assert np.allclose(c1.to_numpy(), np.vstack((v1.to_numpy(), indexed(v2.to_numpy(), ind))))
    except NotImplementedError:
        pass
    c1.append(c2[ind], remove_from_other=True)
    assert len(c2) == len(ind_complement_)
    assert c2.space == c1.space
    assert len(c1) == len_v1 + 2 * len_ind
    assert np.all(almost_equal(c1[len_v1:len_v1 + len_ind], c1[len_v1 + len_ind:len(c1)]))
    assert np.all(almost_equal(c2, v2[ind_complement_]))
    try:
        assert np.allclose(c2.to_numpy(), indexed(v2.to_numpy(), ind_complement_))
    except NotImplementedError:
        pass


@pyst.given_vector_arrays()
def test_append_self(vector_array):
    c = vector_array.copy()
    len_v = len(vector_array)
    c.append(c)
    assert len(c) == 2 * len_v
    assert np.all(almost_equal(c[:len_v], c[len_v:len(c)]))
    try:
        assert np.allclose(c.to_numpy(), np.vstack((vector_array.to_numpy(), vector_array.to_numpy())))
    except NotImplementedError:
        pass
    c = vector_array.copy()
    with pytest.raises(Exception):
        vector_array.append(vector_array, remove_from_other=True)


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices)
def test_del(vectors_and_indices):
    v, ind = vectors_and_indices
    ind_complement_ = ind_complement(v, ind)
    c = v.copy()
    del c[ind]
    assert c.space == v.space
    assert len(c) == len(ind_complement_)
    assert np.all(almost_equal(v[ind_complement_], c))
    try:
        assert np.allclose(c.to_numpy(), indexed(v.to_numpy(), ind_complement_))
    except NotImplementedError:
        pass
    del c[:]
    assert len(c) == 0


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices)
def test_scla(vectors_and_indices):
    v, ind = vectors_and_indices
    if v.len_ind(ind) != v.len_ind_unique(ind):
        with pytest.raises(Exception):
            c = v.copy()
            c[ind].scal(1.)
        return
    ind_complement_ = ind_complement(v, ind)
    c = v.copy()
    c[ind].scal(1.)
    assert len(c) == len(v)
    assert np.all(almost_equal(c, v))

    c = v.copy()
    c[ind].scal(0.)
    assert np.all(almost_equal(c[ind], v.zeros(v.len_ind(ind))))
    assert np.all(almost_equal(c[ind_complement_], v[ind_complement_]))

    for x in (1., 1.4, np.random.random(v.len_ind(ind))):
        c = v.copy()
        c[ind].scal(x)
        assert np.all(almost_equal(c[ind_complement_], v[ind_complement_]))
        assert np.allclose(c[ind].sup_norm(), v[ind].sup_norm() * abs(x))
        assert np.allclose(c[ind].norm(), v[ind].norm() * abs(x))
        try:
            y = v.to_numpy(True)
            if isinstance(x, np.ndarray) and not isinstance(ind, Number):
                x = x[:, np.newaxis]
            y[ind] *= x
            assert np.allclose(c.to_numpy(), y)
        except NotImplementedError:
            pass


@pyst.given_vector_arrays(count=2, index_strategy=pyst.pairs_same_length,
                          scalar=hyst.floats(min_value=1, max_value=pyst.MAX_VECTORARRAY_LENGTH))
def test_axpy(vectors_and_indices, scalar):
    (v1, v2), (ind1, ind2) = vectors_and_indices
    if v1.len_ind(ind1) != v1.len_ind_unique(ind1):
        with pytest.raises(Exception):
            c1, c2 = v1.copy(), v2.copy()
            c1[ind1].axpy(0., c2[ind2])
        return
    # ind2 is used for axpy args
    len_ind2 = v2.len_ind(ind2)
    assume(len_ind2 == 1 or len_ind2 == v1.len_ind(ind1))
    ind1_complement = ind_complement(v1, ind1)
    c1, c2 = v1.copy(), v2.copy()
    c1[ind1].axpy(0., c2[ind2])
    assert len(c1) == len(v1)
    assert np.all(almost_equal(c1, v1))
    assert np.all(almost_equal(c2, v2))

    a = scalar
    c1, c2 = v1.copy(), v2.copy()
    c1[ind1].axpy(a, c2[ind2])
    assert len(c1) == len(v1)
    assert np.all(almost_equal(c1[ind1_complement], v1[ind1_complement]))
    assert np.all(almost_equal(c2, v2))
    assert np.all(c1[ind1].sup_norm() <= v1[ind1].sup_norm() + abs(a) * v2[ind2].sup_norm() * (1. + 1e-10))
    assert np.all(c1[ind1].norm() <= (v1[ind1].norm() + abs(a) * v2[ind2].norm()) * (1. + 1e-10))
    try:
        x = v1.to_numpy(True).astype(complex)  # ensure that inplace addition works
        if isinstance(ind1, Number):
            x[[ind1]] += indexed(v2.to_numpy(), ind2) * a
        else:
            if isinstance(a, np.ndarray):
                aa = a[:, np.newaxis]
            else:
                aa = a
            x[ind1] += indexed(v2.to_numpy(), ind2) * aa
        assert np.allclose(c1.to_numpy(), x)
    except NotImplementedError:
        pass
    c1[ind1].axpy(-a, c2[ind2])
    assert len(c1) == len(v1)
    assert np.all(almost_equal(c1, v1, atol=1e-13, rtol=1e-13))


@pyst.given_vector_arrays(count=2, index_strategy=pyst.pairs_same_length,
                          scalar=hyst.floats(min_value=1, max_value=pyst.MAX_VECTORARRAY_LENGTH))
def test_axpy_one_x(vectors_and_indices, scalar):
    (v1, v2), (ind1, _) = vectors_and_indices
    for ind2 in pyst.valid_inds(v2, 1, random_module=False):
        assert v1.check_ind(ind1)
        assert v2.check_ind(ind2)
        if v1.len_ind(ind1) != v1.len_ind_unique(ind1):
            with pytest.raises(Exception):
                c1, c2 = v1.copy(), v2.copy()
                c1[ind1].axpy(0., c2[ind2])
            continue

        ind1_complement = ind_complement(v1, ind1)
        c1, c2 = v1.copy(), v2.copy()

        gc = c1[ind1]
        gv = c2[ind2]
        gc.axpy(0., gv)
        assert len(c1) == len(v1)
        assert np.all(almost_equal(c1, v1))
        assert np.all(almost_equal(c2, v2))

        a = scalar
        c1, c2 = v1.copy(), v2.copy()
        c1[ind1].axpy(a, c2[ind2])
        assert len(c1) == len(v1)
        assert np.all(almost_equal(c1[ind1_complement], v1[ind1_complement]))
        assert np.all(almost_equal(c2, v2))
        # for the openstack CI machines this could be 1 + 1e-10
        rtol_factor = 1. + 147e-9
        assert np.all(c1[ind1].sup_norm() <= v1[ind1].sup_norm() + abs(a) * v2[ind2].sup_norm() * rtol_factor)
        assert np.all(c1[ind1].norm() <= (v1[ind1].norm() + abs(a) * v2[ind2].norm()) * (1. + 1e-10))
        try:
            x = v1.to_numpy(True).astype(complex)  # ensure that inplace addition works
            if isinstance(ind1, Number):
                x[[ind1]] += indexed(v2.to_numpy(), ind2) * a
            else:
                if isinstance(a, np.ndarray):
                    aa = a[:, np.newaxis]
                else:
                    aa = a
                x[ind1] += indexed(v2.to_numpy(), ind2) * aa
            assert np.allclose(c1.to_numpy(), x)
        except NotImplementedError:
            pass
        c1[ind1].axpy(-a, c2[ind2])
        assert len(c1) == len(v1)
        assert np.all(almost_equal(c1, v1, atol=1e-13, rtol=1e-13))


@pyst.given_vector_arrays(index_strategy=pyst.pairs_same_length,
                          scalar=hyst.floats(min_value=1, max_value=pyst.MAX_VECTORARRAY_LENGTH))
def test_axpy_self(vectors_and_indices, scalar):
    v, (ind1, ind2) = vectors_and_indices
    if v.len_ind(ind1) != v.len_ind_unique(ind1):
        with pytest.raises(Exception):
            c, = v.copy()
            c[ind1].axpy(0., c[ind2])
        return

    ind1_complement = ind_complement(v, ind1)
    c = v.copy()
    rr = c[ind2]
    lp = c[ind1]
    lp.axpy(0., rr)
    assert len(c) == len(v)
    assert np.all(almost_equal(c, v))
    a = scalar
    c = v.copy()
    c[ind1].axpy(a, c[ind2])
    assert len(c) == len(v)
    assert np.all(almost_equal(c[ind1_complement], v[ind1_complement]))
    assert np.all(c[ind1].sup_norm() <= v[ind1].sup_norm() + abs(a) * v[ind2].sup_norm() * (1. + 1e-10))
    try:
        x = v.to_numpy(True).astype(complex)  # ensure that inplace addition works
        if isinstance(ind1, Number):
            x[[ind1]] += indexed(v.to_numpy(), ind2) * a
        else:
            if isinstance(a, np.ndarray):
                aa = a[:, np.newaxis]
            else:
                aa = a
            x[ind1] += indexed(v.to_numpy(), ind2) * aa
        assert np.allclose(c.to_numpy(), x)
    except NotImplementedError:
        pass
    c[ind1].axpy(-a, v[ind2])
    assert len(c) == len(v)
    assert np.all(almost_equal(c, v))

    ind = ind1
    if v.len_ind(ind) != v.len_ind_unique(ind):
        return

    for x in (1., 23., -4):
        c = v.copy()
        cc = v.copy()
        c[ind].axpy(x, c[ind])
        cc[ind].scal(1 + x)
        assert np.all(almost_equal(c, cc))


@pyst.given_vector_arrays(count=2)
def test_pairwise_inner(vector_arrays):
    v1, v2 = vector_arrays
    for ind1, ind2 in pyst.valid_inds_of_same_length(v1, v2):
        r = v1[ind1].pairwise_inner(v2[ind2])
        assert isinstance(r, np.ndarray)
        assert r.shape == (v1.len_ind(ind1),)
        r2 = v2[ind2].pairwise_inner(v1[ind1])
        assert np.allclose, (r, r2)
        assert np.all(r <= v1[ind1].norm() * v2[ind2].norm() * (1. + 1e-10))
        try:
            assert np.allclose(r, np.sum(indexed(v1.to_numpy(), ind1).conj() * indexed(v2.to_numpy(), ind2), axis=1))
        except NotImplementedError:
            pass


@pyst.given_vector_arrays(index_strategy=pyst.pairs_same_length)
def test_pairwise_inner_self(vectors_and_indices):
    v, (ind1, ind2) = vectors_and_indices
    r = v[ind1].pairwise_inner(v[ind2])
    assert isinstance(r, np.ndarray)
    assert r.shape == (v.len_ind(ind1),)
    r2 = v[ind2].pairwise_inner(v[ind1])
    assert np.allclose(r, r2.T.conj())
    assert np.all(r <= v[ind1].norm() * v[ind2].norm() * (1. + 1e-10))
    try:
        assert np.allclose(r, np.sum(indexed(v.to_numpy(), ind1).conj() * indexed(v.to_numpy(), ind2), axis=1))
    except NotImplementedError:
        pass
    ind = ind1
    r = v[ind].pairwise_inner(v[ind])
    assert np.allclose(r, v[ind].norm() ** 2)


@settings(deadline=None, print_blob=True)
@pyst.given_vector_arrays(count=2, index_strategy=pyst.pairs_both_lengths)
def test_inner(vectors_and_indices):
    (v1, v2), (ind1, ind2) = vectors_and_indices
    r = v1[ind1].inner(v2[ind2])
    assert isinstance(r, np.ndarray)
    assert r.shape == (v1.len_ind(ind1), v2.len_ind(ind2))
    r2 = v2[ind2].inner(v1[ind1])
    assert np.allclose(r, r2.T.conj())
    assert np.all(r <= v1[ind1].norm()[:, np.newaxis] * v2[ind2].norm()[np.newaxis, :] * (1. + 1e-10))
    try:
        assert np.allclose(r, indexed(v1.to_numpy(), ind1).conj().dot(indexed(v2.to_numpy(), ind2).T))
    except NotImplementedError:
        pass


@settings(deadline=None)
@pyst.given_vector_arrays(index_strategy=pyst.pairs_both_lengths)
def test_inner_self(vectors_and_indices):
    v, (ind1, ind2) = vectors_and_indices
    r = v[ind1].inner(v[ind2])
    assert isinstance(r, np.ndarray)
    assert r.shape == (v.len_ind(ind1), v.len_ind(ind2))
    r2 = v[ind2].inner(v[ind1])
    assert np.allclose(r, r2.T.conj())
    assert np.all(r <= v[ind1].norm()[:, np.newaxis] * v[ind2].norm()[np.newaxis, :] * (1. + 1e-10))
    try:
        assert np.allclose(r, indexed(v.to_numpy(), ind1).conj().dot(indexed(v.to_numpy(), ind2).T))
    except NotImplementedError:
        pass
    r = v[ind1].inner(v[ind1])
    assert np.allclose(r, r.T.conj())


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices, random=hyst.random_module())
def test_lincomb_1d(vectors_and_indices, random):
    v, ind = vectors_and_indices
    coeffs = np.random.random(v.len_ind(ind))
    lc = v[ind].lincomb(coeffs)
    assert lc.space == v.space
    assert len(lc) == 1
    lc2 = v.zeros()
    for coeff, i in zip(coeffs, ind_to_list(v, ind)):
        lc2.axpy(coeff, v[i])
    assert np.all(almost_equal(lc, lc2))


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices, random=hyst.random_module())
def test_lincomb_2d(vectors_and_indices, random):
    v, ind = vectors_and_indices
    for count in (0, 1, 5):
        coeffs = np.random.random((count, v.len_ind(ind)))
        lc = v[ind].lincomb(coeffs)
        assert lc.space == v.space
        assert len(lc) == count
        lc2 = v.empty(reserve=count)
        for coeffs_1d in coeffs:
            lc2.append(v[ind].lincomb(coeffs_1d))
        assert np.all(almost_equal(lc, lc2))


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices, random=hyst.random_module())
def test_lincomb_wrong_coefficients(vectors_and_indices, random):
    v, ind = vectors_and_indices
    coeffs = np.random.random(v.len_ind(ind) + 1)
    with pytest.raises(Exception):
        v[ind].lincomb(coeffs)
    coeffs = np.random.random(v.len_ind(ind)).reshape((1, 1, -1))
    with pytest.raises(Exception):
        v[ind].lincomb(coeffs)
    if v.len_ind(ind) > 0:
        coeffs = np.random.random(v.len_ind(ind) - 1)
        with pytest.raises(Exception):
            v[ind].lincomb(coeffs)
        coeffs = np.array([])
        with pytest.raises(Exception):
            v[ind].lincomb(coeffs)


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices)
def test_norm(vectors_and_indices):
    v, ind = vectors_and_indices
    c = v.copy()
    norm = c[ind].norm()
    assert isinstance(norm, np.ndarray)
    assert norm.shape == (v.len_ind(ind),)
    assert np.all(norm >= 0)
    if v.dim == 0:
        assert np.all(norm == 0)
    try:
        assert np.allclose(norm, np.linalg.norm(indexed(v.to_numpy(), ind), axis=1))
    except NotImplementedError:
        pass
    c.scal(4.)
    assert np.allclose(c[ind].norm(), norm * 4)
    c.scal(-4.)
    assert np.allclose(c[ind].norm(), norm * 16)
    c.scal(0.)
    assert np.allclose(c[ind].norm(), 0)


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices)
def test_norm2(vectors_and_indices):
    v, ind = vectors_and_indices
    c = v.copy()
    norm = c[ind].norm2()
    assert isinstance(norm, np.ndarray)
    assert norm.shape == (v.len_ind(ind),)
    assert np.all(norm >= 0)
    if v.dim == 0:
        assert np.all(norm == 0)
    try:
        assert np.allclose(norm, np.linalg.norm(indexed(v.to_numpy(), ind), axis=1)**2)
    except NotImplementedError:
        pass
    c.scal(4.)
    assert np.allclose(c[ind].norm2(), norm * 16)
    c.scal(-4.)
    assert np.allclose(c[ind].norm2(), norm * 256)
    c.scal(0.)
    assert np.allclose(c[ind].norm2(), 0)


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices)
def test_sup_norm(vectors_and_indices):
    v, ind = vectors_and_indices
    c = v.copy()
    norm = c[ind].sup_norm()
    assert isinstance(norm, np.ndarray)
    assert norm.shape == (v.len_ind(ind),)
    assert np.all(norm >= 0)
    if v.dim == 0:
        assert np.all(norm == 0)
    if v.dim > 0:
        try:
            assert np.allclose(norm, np.max(np.abs(indexed(v.to_numpy(), ind)), axis=1))
        except NotImplementedError:
            pass
    c.scal(4.)
    assert np.allclose(c[ind].sup_norm(), norm * 4)
    c.scal(-4.)
    assert np.allclose(c[ind].sup_norm(), norm * 16)
    c.scal(0.)
    assert np.allclose(c[ind].sup_norm(), 0)


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices, random_count=hyst.integers(min_value=1, max_value=10))
def test_dofs(vectors_and_indices, random_count):
    v, ind = vectors_and_indices
    c = v.copy()
    dofs = c[ind].dofs(np.array([], dtype=int))
    assert isinstance(dofs, np.ndarray)
    assert dofs.shape == (v.len_ind(ind), 0)

    c = v.copy()
    dofs = c[ind].dofs([])
    assert isinstance(dofs, np.ndarray)
    assert dofs.shape == (v.len_ind(ind), 0)

    assume(v.dim > 0)

    c_ind = np.random.randint(0, v.dim, random_count)
    c = v.copy()
    dofs = c[ind].dofs(c_ind)
    assert dofs.shape == (v.len_ind(ind), random_count)
    c = v.copy()
    dofs2 = c[ind].dofs(list(c_ind))
    assert np.all(dofs == dofs2)
    c = v.copy()
    c.scal(3.)
    dofs2 = c[ind].dofs(c_ind)
    assert np.allclose(dofs * 3, dofs2)
    c = v.copy()
    dofs2 = c[ind].dofs(np.hstack((c_ind, c_ind)))
    assert np.all(dofs2 == np.hstack((dofs, dofs)))
    try:
        assert np.all(dofs == indexed(v.to_numpy(), ind)[:, c_ind])
    except NotImplementedError:
        pass


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices)
def test_components_wrong_dof_indices(vectors_and_indices):
    v, ind = vectors_and_indices
    with pytest.raises(Exception):
        v[ind].dofs(None)
    with pytest.raises(Exception):
        v[ind].dofs(1)
    with pytest.raises(Exception):
        v[ind].dofs(np.array([-1]))
    with pytest.raises(Exception):
        v[ind].dofs(np.array([v.dim]))


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices)
def test_amax(vectors_and_indices):
    v, ind = vectors_and_indices
    assume(v.dim > 0)
    max_inds, max_vals = v[ind].amax()
    assert np.allclose(max_vals, v[ind].sup_norm())
    for i, max_ind, max_val in zip(ind_to_list(v, ind), max_inds, max_vals):
        assert np.allclose(max_val, np.abs(v[[i]].dofs([max_ind])))


# def test_amax_zero_dim(zero_dimensional_vector_space):
#     for count in (0, 10):
#         v = zero_dimensional_vector_space.zeros(count=count)
#         for ind in valid_inds(v):
#             with pytest.raises(Exception):
#                 v.amax(ind)


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices)
def test_gramian(vectors_and_indices):
    v, ind = vectors_and_indices
    assert np.allclose(v[ind].gramian(), v[ind].inner(v[ind]))


@pyst.given_vector_arrays(count=2, length=pyst.equal_tuples(pyst.hy_lengths, count=2))
def test_add(vector_arrays):
    v1, v2 = vector_arrays
    c1 = v1.copy()
    cc1 = v1.copy()
    c1.axpy(1, v2)
    assert np.all(almost_equal(v1 + v2, c1))
    assert np.all(almost_equal(v1, cc1))


@pyst.given_vector_arrays(count=2, length=pyst.equal_tuples(pyst.hy_lengths, count=2))
def test_iadd(vector_arrays):
    v1, v2 = vector_arrays
    c1 = v1.copy()
    c1.axpy(1, v2)
    v1 += v2
    assert np.all(almost_equal(v1, c1))


@pyst.given_vector_arrays(count=2, length=pyst.equal_tuples(pyst.hy_lengths, count=2))
def test_sub(vector_arrays):
    v1, v2 = vector_arrays
    c1 = v1.copy()
    cc1 = v1.copy()
    c1.axpy(-1, v2)
    assert np.all(almost_equal((v1 - v2), c1))
    assert np.all(almost_equal(v1, cc1))


@pyst.given_vector_arrays(count=2, length=pyst.equal_tuples(pyst.hy_lengths, count=2))
def test_isub(vector_arrays):
    v1, v2 = vector_arrays
    c1 = v1.copy()
    c1.axpy(-1, v2)
    v1 -= v2
    assert np.all(almost_equal(v1, c1))


@pyst.given_vector_arrays()
def test_neg(vector_array):
    c = vector_array.copy()
    cc = vector_array.copy()
    c.scal(-1)
    assert np.all(almost_equal(c, -vector_array))
    assert np.all(almost_equal(vector_array, cc))


@pyst.given_vector_arrays(index_strategy=pyst.st_scaling_value)
def test_mul(vectors_and_indices):
    vector_array, a = vectors_and_indices
    c = vector_array.copy()
    cc = vector_array.copy()
    cc.scal(a)
    assert np.all(almost_equal((vector_array * a), cc))
    assert np.all(almost_equal(vector_array, c))


@pyst.given_vector_arrays()
def test_mul_wrong_factor(vector_array):
    with pytest.raises(Exception):
        _ = vector_array * vector_array


@pyst.given_vector_arrays(index_strategy=pyst.st_scaling_value)
def test_rmul(vectors_and_indices):
    vector_array, a = vectors_and_indices
    c = vector_array.copy()
    cc = vector_array.copy()
    cc.scal(a)
    alpha = a * vector_array
    # the scaling_value strategy also draws ndarrays, for which alpha here will be an ndarray,
    # which in turn will fail the axpy hidden in the almost_equal check
    assume(not isinstance(alpha, np.ndarray))
    assert np.all(almost_equal(alpha, cc))
    assert np.all(almost_equal(vector_array, c))


@pyst.given_vector_arrays(index_strategy=pyst.st_scaling_value)
def test_imul(vectors_and_indices):
    vector_array, a = vectors_and_indices
    c = vector_array.copy()
    cc = vector_array.copy()
    c.scal(a)
    cc *= a
    assert np.all(almost_equal(c, cc))


@pyst.given_vector_arrays()
def test_imul_wrong_factor(vector_array):
    with pytest.raises(Exception):
        vector_array *= vector_array


@pyst.given_vector_arrays()
def test_iter(vector_array):
    v = vector_array
    w = v.empty()
    for vv in v:
        w.append(vv)
    assert np.all(almost_equal(w, v))


####################################################################################################


@pyst.given_vector_arrays(count=2, compatible=False)
def test_append_incompatible(vector_arrays):
    v1, v2 = vector_arrays
    c1, c2 = v1.copy(), v2.copy()
    with pytest.raises(Exception):
        c1.append(c2, remove_from_other=False)
    c1, c2 = v1.copy(), v2.copy()
    with pytest.raises(Exception):
        c1.append(c2, remove_from_other=True)


@pyst.given_vector_arrays(count=2, compatible=False)
def test_axpy_incompatible(vector_arrays):
    v1, v2 = vector_arrays
    for ind1, ind2 in pyst.valid_inds_of_same_length(v1, v2, random_module=False):
        c1, c2 = v1.copy(), v2.copy()
        with pytest.raises(Exception):
            c1[ind1].axpy(0., c2[ind2])
        c1, c2 = v1.copy(), v2.copy()
        with pytest.raises(Exception):
            c1[ind1].axpy(1., c2[ind2])
        c1, c2 = v1.copy(), v2.copy()
        with pytest.raises(Exception):
            c1[ind1].axpy(-1., c2[ind2])
        c1, c2 = v1.copy(), v2.copy()
        with pytest.raises(Exception):
            c1[ind1].axpy(1.42, c2[ind2])


@pyst.given_vector_arrays(count=2, compatible=False)
def test_inner_incompatible(vector_arrays):
    v1, v2 = vector_arrays
    for ind1, ind2 in pyst.valid_inds_of_same_length(v1, v2, random_module=False):
        c1, c2 = v1.copy(), v2.copy()
        with pytest.raises(Exception):
            c1[ind1].inner(c2[ind2])


@pyst.given_vector_arrays(count=2, compatible=False)
def test_pairwise_inner_incompatible(vector_arrays):
    v1, v2 = vector_arrays
    for ind1, ind2 in pyst.valid_inds_of_same_length(v1, v2, random_module=False):
        c1, c2 = v1.copy(), v2.copy()
        with pytest.raises(Exception):
            c1[ind1].pairwise_inner(c2[ind2])


@pyst.given_vector_arrays(count=2, compatible=False)
def test_add_incompatible(vector_arrays):
    v1, v2 = vector_arrays
    with pytest.raises(Exception):
        _ = v1 + v2


@pyst.given_vector_arrays(count=2, compatible=False)
def test_iadd_incompatible(vector_arrays):
    v1, v2 = vector_arrays
    with pytest.raises(Exception):
        v1 += v2


@pyst.given_vector_arrays(count=2, compatible=False)
def test_sub_incompatible(vector_arrays):
    v1, v2 = vector_arrays
    with pytest.raises(Exception):
        _ = v1 - v2


@pyst.given_vector_arrays(count=2, compatible=False)
def test_isub_incompatible(vector_arrays):
    v1, v2 = vector_arrays
    with pytest.raises(Exception):
        v1 -= v2


####################################################################################################


@pyst.given_vector_arrays(index_strategy=pyst.invalid_indices)
def test_wrong_ind_raises_exception(vectors_and_indices):
    vector_array, ind = vectors_and_indices
    with pytest.raises(Exception):
        vector_array[ind]


@pyst.given_vector_arrays(index_strategy=pyst.valid_indices)
def test_scal_wrong_coefficients(vectors_and_indices):
    v, ind = vectors_and_indices
    for alpha in ([np.array([]), np.eye(v.len_ind(ind)), np.random.random(v.len_ind(ind) + 1)]
                  if v.len_ind(ind) > 0 else
                  [np.random.random(1)]):
        with pytest.raises(Exception):
            v[ind].scal(alpha)


@pyst.given_vector_arrays(count=2, index_strategy=pyst.pairs_same_length)
def test_axpy_wrong_coefficients(vectors_and_indices):
    (v1, v2), (ind1, ind2) = vectors_and_indices
    for alpha in ([np.array([]), np.eye(v1.len_ind(ind1)), np.random.random(v1.len_ind(ind1) + 1)]
                  if v1.len_ind(ind1) > 0 else
                  [np.random.random(1)]):
        with pytest.raises(Exception):
            v1[ind1].axpy(alpha, v2[ind2])


@pyst.given_vector_arrays(which='picklable')
def test_pickle(vector_array):
    assert_picklable_without_dumps_function(vector_array)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

"""Reductors based on H2-norm."""

from numbers import Integral, Real

import numpy as np
import scipy.linalg as spla

from pymor.algorithms.gram_schmidt import gram_schmidt, gram_schmidt_biorth
from pymor.algorithms.krylov import tangential_rational_krylov
from pymor.algorithms.riccati import solve_ricc_dense
from pymor.algorithms.sylvester import solve_sylv_schur
from pymor.algorithms.to_matrix import to_matrix
from pymor.core.base import BasicObject
from pymor.models.iosys import InputOutputModel, LTIModel, _lti_to_poles_b_c, _poles_b_c_to_lti
from pymor.operators.constructions import IdentityOperator
from pymor.parameters.base import Mu
from pymor.reductors.basic import LTIPGReductor
from pymor.reductors.interpolation import LTIBHIReductor, TFBHIReductor


class GenericIRKAReductor(BasicObject):
    """Generic IRKA related reductor.

    Parameters
    ----------
    fom
        The full-order |Model| to reduce.
    mu
        |Parameter values|.
    """

    def _clear_lists(self):
        self.sigma_list = []
        self.b_list = []
        self.c_list = []
        self.conv_crit = []
        self._conv_data = []
        self.errors = []

    def __init__(self, fom, mu=None):
        if not isinstance(mu, Mu):
            mu = fom.parameters.parse(mu)
        assert fom.parameters.assert_compatible(mu)
        self.fom = fom
        self.mu = mu
        self.V = None
        self.W = None
        self._pg_reductor = None
        self._clear_lists()

    def reconstruct(self, u):
        """Reconstruct high-dimensional vector from reduced vector `u`."""
        return self._pg_reductor.reconstruct(u)

    def _check_rom0_params(self, rom0_params):
        if isinstance(rom0_params, Integral):
            assert rom0_params > 0
            if hasattr(self.fom, 'order'):  # self.fom can be a TransferFunction
                assert rom0_params < self.fom.order
        elif isinstance(rom0_params, np.ndarray):
            assert rom0_params.ndim == 1
        elif isinstance(rom0_params, dict):
            assert ('sigma', 'b', 'c') in rom0_params
            assert isinstance(rom0_params['sigma'], np.ndarray)
            assert isinstance(rom0_params['b'], np.ndarray)
            assert isinstance(rom0_params['c'], np.ndarray)
            assert rom0_params['sigma'].ndim == 1
            assert rom0_params['b'].shape[1] == self.fom.dim_input
            assert rom0_params['c'].shape[1] == self.fom.dim_output
            assert len(rom0_params['sigma']) == rom0_params['b'].shape[0]
            assert len(rom0_params['sigma']) == rom0_params['c'].shape[0]
        elif isinstance(rom0_params, LTIModel):
            assert rom0_params.order > 0
            if hasattr(self.fom, 'order'):  # self.fom can be a TransferFunction
                assert rom0_params < self.fom.order
            assert rom0_params.dim_input == self.fom.dim_input
            assert rom0_params.dim_output == self.fom.dim_output
        else:
            raise ValueError(f'rom0_params is of wrong type ({type(rom0_params)}).')

    @staticmethod
    def _check_common_args(tol, maxit, num_prev, conv_crit):
        assert isinstance(tol, Real) and tol > 0
        assert isinstance(maxit, Integral) and maxit >= 1
        assert isinstance(num_prev, Integral) and num_prev >= 1
        assert conv_crit in ('sigma', 'h2')

    def _order_to_sigma_b_c(self, r):
        sigma = np.logspace(-1, 1, r)
        b = (np.ones((r, 1))
             if self.fom.dim_input == 1
             else np.random.RandomState(0).normal(size=(r, self.fom.dim_input)))
        c = (np.ones((r, 1))
             if self.fom.dim_output == 1
             else np.random.RandomState(0).normal(size=(r, self.fom.dim_output)))
        return sigma, b, c

    @staticmethod
    def _rom_to_sigma_b_c(rom, force_sigma_in_rhp):
        poles, b, c = _lti_to_poles_b_c(rom)
        sigma = (np.abs(poles.real) + poles.imag * 1j
                 if force_sigma_in_rhp
                 else -poles)
        return sigma, b, c

    def _rom0_params_to_sigma_b_c(self, rom0_params, force_sigma_in_rhp):
        self.logger.info('Generating initial interpolation data')
        self._check_rom0_params(rom0_params)
        if isinstance(rom0_params, Integral):
            sigma, b, c = self._order_to_sigma_b_c(rom0_params)
        elif isinstance(rom0_params, np.ndarray):
            sigma = rom0_params
            _, b, c = self._order_to_sigma_b_c(len(rom0_params))
        elif isinstance(rom0_params, dict):
            sigma = rom0_params['sigma']
            b = rom0_params['b']
            c = rom0_params['c']
        else:
            sigma, b, c = self._rom_to_sigma_b_c(rom0_params, force_sigma_in_rhp)
        return sigma, b, c

    def _rom0_params_to_rom(self, rom0_params):
        self.logger.info('Generating initial reduced-order model')
        self._check_rom0_params(rom0_params)
        if isinstance(rom0_params, Integral):
            sigma, b, c = self._order_to_sigma_b_c(rom0_params)
            rom0 = _poles_b_c_to_lti(-sigma, b, c)
        elif isinstance(rom0_params, np.ndarray):
            sigma = rom0_params
            _, b, c = self._order_to_sigma_b_c(len(rom0_params))
            rom0 = _poles_b_c_to_lti(-sigma, b, c)
        elif isinstance(rom0_params, dict):
            sigma = rom0_params['sigma']
            b = rom0_params['b']
            c = rom0_params['c']
            rom0 = _poles_b_c_to_lti(-sigma, b, c)
        else:
            rom0 = rom0_params
        return rom0

    def _store_sigma_b_c(self, sigma, b, c):
        if sigma is not None:
            self.sigma_list.append(sigma)
        if b is not None:
            self.b_list.append(b)
        if c is not None:
            self.c_list.append(c)

    def _update_conv_data(self, sigma, rom, conv_crit):
        del self._conv_data[-1]
        self._conv_data.insert(0, sigma if conv_crit == 'sigma' else rom)

    def _compute_conv_crit(self, rom, conv_crit, it):
        if conv_crit == 'sigma':
            sigma = self._conv_data[0]
            dist = min(spla.norm((sigma_old - sigma) / sigma_old, ord=np.inf)
                       for sigma_old in self._conv_data[1:]
                       if sigma_old is not None)
        else:
            if rom.poles().real.max() >= 0:
                dist = np.inf
            else:
                dist = min((rom_old - rom).h2_norm() / rom_old.h2_norm()
                           if rom_old is not None and rom_old.poles().real.max() < 0
                           else np.inf
                           for rom_old in self._conv_data[1:])
        self.conv_crit.append(dist)
        self.logger.info(f'Convergence criterion in iteration {it + 1}: {dist:e}')

    def _compute_error(self, rom, it, compute_errors):
        if not compute_errors:
            return
        rel_h2_err = ((self.fom - rom).h2_norm() / self.fom.h2_norm()
                      if rom.poles().real.max() < 0
                      else np.inf)
        self.errors.append(rel_h2_err)
        self.logger.info(f'Relative H2-error in iteration {it + 1}: {rel_h2_err:e}')


class IRKAReductor(GenericIRKAReductor):
    """Iterative Rational Krylov Algorithm reductor.

    Parameters
    ----------
    fom
        The full-order |LTIModel| to reduce.
    mu
        |Parameter values|.
    """

    def __init__(self, fom, mu=None):
        assert isinstance(fom, LTIModel)
        super().__init__(fom, mu=mu)

    def reduce(self, rom0_params, tol=1e-4, maxit=100, num_prev=1,
               force_sigma_in_rhp=False, projection='orth', conv_crit='sigma',
               compute_errors=False):
        r"""Reduce using IRKA.

        See :cite:`GAB08` (Algorithm 4.1) and :cite:`ABG10` (Algorithm 1).

        Parameters
        ----------
        rom0_params
            Can be:

            - order of the reduced model (a positive integer),
            - initial interpolation points (a 1D |NumPy array|),
            - dict with `'sigma'`, `'b'`, `'c'` as keys mapping to
              initial interpolation points (a 1D |NumPy array|), right
              tangential directions (|NumPy array| of shape
              `(len(sigma), fom.dim_input)`), and left tangential directions
              (|NumPy array| of shape `(len(sigma), fom.dim_input)`),
            - initial reduced-order model (|LTIModel|).

            If the order of reduced model is given, initial
            interpolation data is generated randomly.
        tol
            Tolerance for the convergence criterion.
        maxit
            Maximum number of iterations.
        num_prev
            Number of previous iterations to compare the current
            iteration to. Larger number can avoid occasional cyclic
            behavior of IRKA.
        force_sigma_in_rhp
            If `False`, new interpolation are reflections of the current
            reduced-order model's poles. Otherwise, only poles in the
            left half-plane are reflected.
        projection
            Projection method:

            - `'orth'`: projection matrices are orthogonalized with
              respect to the Euclidean inner product
            - `'biorth'`: projection matrices are biorthogolized with
              respect to the E product
            - `'arnoldi'`: projection matrices are orthogonalized using
              the Arnoldi process (available only for SISO systems).
        conv_crit
            Convergence criterion:

            - `'sigma'`: relative change in interpolation points
            - `'h2'`: relative :math:`\mathcal{H}_2` distance of
              reduced-order models
        compute_errors
            Should the relative :math:`\mathcal{H}_2`-errors of
            intermediate reduced-order models be computed.

            .. warning::
                Computing :math:`\mathcal{H}_2`-errors is expensive. Use
                this option only if necessary.

        Returns
        -------
        rom
            Reduced |LTIModel| model.
        """
        if not self.fom.cont_time:
            raise NotImplementedError

        self._clear_lists()
        sigma, b, c = self._rom0_params_to_sigma_b_c(rom0_params, force_sigma_in_rhp)
        self._store_sigma_b_c(sigma, b, c)
        self._check_common_args(tol, maxit, num_prev, conv_crit)
        assert projection in ('orth', 'biorth', 'arnoldi')
        if projection == 'arnoldi':
            assert self.fom.dim_input == self.fom.dim_output == 1

        self.logger.info('Starting IRKA')
        self._conv_data = (num_prev + 1) * [None]
        if conv_crit == 'sigma':
            self._conv_data[0] = sigma
        self._pg_reductor = LTIBHIReductor(self.fom, mu=self.mu)
        for it in range(maxit):
            rom = self._pg_reductor.reduce(sigma, b, c, projection=projection)
            sigma, b, c = self._rom_to_sigma_b_c(rom, force_sigma_in_rhp)
            self._store_sigma_b_c(sigma, b, c)
            self._update_conv_data(sigma, rom, conv_crit)
            self._compute_conv_crit(rom, conv_crit, it)
            self._compute_error(rom, it, compute_errors)
            if self.conv_crit[-1] < tol:
                break

        self.V = self._pg_reductor.V
        self.W = self._pg_reductor.W
        return rom


class OneSidedIRKAReductor(GenericIRKAReductor):
    """One-Sided Iterative Rational Krylov Algorithm reductor.

    Parameters
    ----------
    fom
        The full-order |LTIModel| to reduce.
    version
        Version of the one-sided IRKA:

        - `'V'`: Galerkin projection using the input Krylov subspace,
        - `'W'`: Galerkin projection using the output Krylov subspace.
    mu
        |Parameter values|.
    """

    def __init__(self, fom, version, mu=None):
        assert isinstance(fom, LTIModel)
        assert version in ('V', 'W')
        super().__init__(fom, mu=mu)
        self.version = version

    def reduce(self, rom0_params, tol=1e-4, maxit=100, num_prev=1,
               force_sigma_in_rhp=False, projection='orth', conv_crit='sigma',
               compute_errors=False):
        r"""Reduce using one-sided IRKA.

        Parameters
        ----------
        rom0_params
            Can be:

            - order of the reduced model (a positive integer),
            - initial interpolation points (a 1D |NumPy array|),
            - dict with `'sigma'`, `'b'`, `'c'` as keys mapping to
              initial interpolation points (a 1D |NumPy array|), right
              tangential directions (|NumPy array| of shape
              `(len(sigma), fom.dim_input)`), and left tangential directions
              (|NumPy array| of shape `(len(sigma), fom.dim_input)`),
            - initial reduced-order model (|LTIModel|).

            If the order of reduced model is given, initial
            interpolation data is generated randomly.
        tol
            Tolerance for the largest change in interpolation points.
        maxit
            Maximum number of iterations.
        num_prev
            Number of previous iterations to compare the current
            iteration to. A larger number can avoid occasional cyclic
            behavior.
        force_sigma_in_rhp
            If `False`, new interpolation are reflections of the current
            reduced-order model's poles. Otherwise, only poles in the
            left half-plane are reflected.
        projection
            Projection method:

            - `'orth'`: projection matrix is orthogonalized with respect
              to the Euclidean inner product,
            - `'Eorth'`: projection matrix is orthogonalized with
              respect to the E product.
        conv_crit
            Convergence criterion:

            - `'sigma'`: relative change in interpolation points,
            - `'h2'`: relative :math:`\mathcal{H}_2` distance of
              reduced-order models.
        compute_errors
            Should the relative :math:`\mathcal{H}_2`-errors of
            intermediate reduced-order models be computed.

            .. warning::
                Computing :math:`\mathcal{H}_2`-errors is expensive. Use
                this option only if necessary.

        Returns
        -------
        rom
            Reduced |LTIModel| model.
        """
        if not self.fom.cont_time:
            raise NotImplementedError

        self._clear_lists()
        sigma, b, c = self._rom0_params_to_sigma_b_c(rom0_params, force_sigma_in_rhp)
        self._store_sigma_b_c(sigma, b, c)
        self._check_common_args(tol, maxit, num_prev, conv_crit)
        assert projection in ('orth', 'Eorth')

        self.logger.info('Starting one-sided IRKA')
        self._conv_data = (num_prev + 1) * [None]
        if conv_crit == 'sigma':
            self._conv_data[0] = sigma
        for it in range(maxit):
            self._set_V_reductor(sigma, b, c, projection)
            rom = self._pg_reductor.reduce()
            sigma, b, c = self._rom_to_sigma_b_c(rom, force_sigma_in_rhp)
            self._store_sigma_b_c(sigma, b, c)
            self._update_conv_data(sigma, rom, conv_crit)
            self._compute_conv_crit(rom, conv_crit, it)
            self._compute_error(rom, it, compute_errors)
            if self.conv_crit[-1] < tol:
                break

        return rom

    def _set_V_reductor(self, sigma, b, c, projection):
        fom = (
            self.fom.with_(
                **{op: getattr(self.fom, op).assemble(mu=self.mu)
                   for op in ['A', 'B', 'C', 'D', 'E']}
            )
            if self.fom.parametric
            else self.fom
        )
        if self.version == 'V':
            self.V = tangential_rational_krylov(fom.A, fom.E, fom.B, fom.B.source.from_numpy(b), sigma,
                                                orth=False)
            gram_schmidt(self.V, atol=0, rtol=0,
                         product=None if projection == 'orth' else fom.E,
                         copy=False)
        else:
            self.V = tangential_rational_krylov(fom.A, fom.E, fom.C, fom.C.range.from_numpy(c), sigma, trans=True,
                                                orth=False)
            gram_schmidt(self.V, atol=0, rtol=0,
                         product=None if projection == 'orth' else fom.E,
                         copy=False)
        self.W = self.V
        self._pg_reductor = LTIPGReductor(fom, self.V, self.V,
                                          projection == 'Eorth')


class TSIAReductor(GenericIRKAReductor):
    """Two-Sided Iteration Algorithm reductor.

    Parameters
    ----------
    fom
        The full-order |LTIModel| to reduce.
    mu
        |Parameter values|.
    """

    def __init__(self, fom, mu=None):
        assert isinstance(fom, LTIModel)
        super().__init__(fom, mu=mu)

    def reduce(self, rom0_params, tol=1e-4, maxit=100, num_prev=1, projection='orth',
               conv_crit='sigma', compute_errors=False):
        r"""Reduce using TSIA.

        See :cite:`XZ11` (Algorithm 1) and :cite:`BKS11`.

        In exact arithmetic, TSIA is equivalent to IRKA (under some
        assumptions on the poles of the reduced model). The main
        difference in implementation is that TSIA computes the Schur
        decomposition of the reduced matrices, while IRKA computes the
        eigenvalue decomposition. Therefore, TSIA might behave better
        for non-normal reduced matrices.

        Parameters
        ----------
        rom0_params
            Can be:

            - order of the reduced model (a positive integer),
            - initial interpolation points (a 1D |NumPy array|),
            - dict with `'sigma'`, `'b'`, `'c'` as keys mapping to
              initial interpolation points (a 1D |NumPy array|), right
              tangential directions (|NumPy array| of shape
              `(len(sigma), fom.dim_input)`), and left tangential directions
              (|NumPy array| of shape `(len(sigma), fom.dim_input)`),
            - initial reduced-order model (|LTIModel|).

            If the order of reduced model is given, initial
            interpolation data is generated randomly.
        tol
            Tolerance for the convergence criterion.
        maxit
            Maximum number of iterations.
        num_prev
            Number of previous iterations to compare the current
            iteration to. Larger number can avoid occasional cyclic
            behavior of TSIA.
        projection
            Projection method:

            - `'orth'`: projection matrices are orthogonalized with
              respect to the Euclidean inner product
            - `'biorth'`: projection matrices are biorthogolized with
              respect to the E product
        conv_crit
            Convergence criterion:

            - `'sigma'`: relative change in interpolation points
            - `'h2'`: relative :math:`\mathcal{H}_2` distance of
              reduced-order models
        compute_errors
            Should the relative :math:`\mathcal{H}_2`-errors of
            intermediate reduced-order models be computed.

            .. warning::
                Computing :math:`\mathcal{H}_2`-errors is expensive. Use
                this option only if necessary.

        Returns
        -------
        rom
            Reduced |LTIModel|.
        """
        if not self.fom.cont_time:
            raise NotImplementedError

        self._clear_lists()
        rom = self._rom0_params_to_rom(rom0_params)
        self._check_common_args(tol, maxit, num_prev, conv_crit)
        assert projection in ('orth', 'biorth')

        self.logger.info('Starting TSIA')
        self._conv_data = (num_prev + 1) * [None]
        self._conv_data[0] = -rom.poles() if conv_crit == 'sigma' else rom
        self._store_sigma_b_c(-rom.poles(), None, None)
        for it in range(maxit):
            self._set_V_W_reductor(rom, projection)
            rom = self._pg_reductor.reduce()
            self._store_sigma_b_c(-rom.poles(), None, None)
            self._update_conv_data(-rom.poles(), rom, conv_crit)
            self._compute_conv_crit(rom, conv_crit, it)
            self._compute_error(rom, it, compute_errors)
            if self.conv_crit[-1] < tol:
                break

        return rom

    def _set_V_W_reductor(self, rom, projection):
        fom = (
            self.fom.with_(
                **{op: getattr(self.fom, op).assemble(mu=self.mu)
                   for op in ['A', 'B', 'C', 'D', 'E']}
            )
            if self.fom.parametric
            else self.fom
        )
        self.V, self.W = solve_sylv_schur(fom.A, rom.A,
                                          E=fom.E, Er=rom.E,
                                          B=fom.B, Br=rom.B,
                                          C=fom.C, Cr=rom.C)
        if projection == 'orth':
            gram_schmidt(self.V, atol=0, rtol=0, copy=False)
            gram_schmidt(self.W, atol=0, rtol=0, copy=False)
        elif projection == 'biorth':
            gram_schmidt_biorth(self.V, self.W, product=fom.E, copy=False)
        self._pg_reductor = LTIPGReductor(fom, self.W, self.V,
                                          projection == 'biorth')


class TFIRKAReductor(GenericIRKAReductor):
    """Realization-independent IRKA reductor.

    See :cite:`BG12`.

    Parameters
    ----------
    fom
        The full-order |Model| with `eval_tf` and `eval_dtf` methods.
    mu
        |Parameter values|.
    """

    def __init__(self, fom, mu=None):
        assert isinstance(fom, InputOutputModel)
        super().__init__(fom, mu=mu)

    def reduce(self, rom0_params, tol=1e-4, maxit=100, num_prev=1,
               force_sigma_in_rhp=False, conv_crit='sigma', compute_errors=False):
        r"""Reduce using TF-IRKA.

        Parameters
        ----------
        rom0_params
            Can be:

            - order of the reduced model (a positive integer),
            - initial interpolation points (a 1D |NumPy array|),
            - dict with `'sigma'`, `'b'`, `'c'` as keys mapping to
              initial interpolation points (a 1D |NumPy array|), right
              tangential directions (|NumPy array| of shape
              `(len(sigma), fom.dim_input)`), and left tangential directions
              (|NumPy array| of shape `(len(sigma), fom.dim_input)`),
            - initial reduced-order model (|LTIModel|).

            If the order of reduced model is given, initial
            interpolation data is generated randomly.
        tol
            Tolerance for the convergence criterion.
        maxit
            Maximum number of iterations.
        num_prev
            Number of previous iterations to compare the current
            iteration to. Larger number can avoid occasional cyclic
            behavior of TF-IRKA.
        force_sigma_in_rhp
            If `False`, new interpolation are reflections of the current
            reduced-order model's poles. Otherwise, only poles in the
            left half-plane are reflected.
        conv_crit
            Convergence criterion:

            - `'sigma'`: relative change in interpolation points
            - `'h2'`: relative :math:`\mathcal{H}_2` distance of
              reduced-order models
        compute_errors
            Should the relative :math:`\mathcal{H}_2`-errors of
            intermediate reduced-order models be computed.

            .. warning::
                Computing :math:`\mathcal{H}_2`-errors is expensive. Use
                this option only if necessary.

        Returns
        -------
        rom
            Reduced |LTIModel| model.
        """
        if not self.fom.cont_time:
            raise NotImplementedError

        self._clear_lists()
        sigma, b, c = self._rom0_params_to_sigma_b_c(rom0_params, force_sigma_in_rhp)
        self._store_sigma_b_c(sigma, b, c)
        self._check_common_args(tol, maxit, num_prev, conv_crit)

        self.logger.info('Starting TF-IRKA')
        self._conv_data = (num_prev + 1) * [None]
        if conv_crit == 'sigma':
            self._conv_data[0] = sigma
        interp_reductor = TFBHIReductor(self.fom, mu=self.mu)
        for it in range(maxit):
            rom = interp_reductor.reduce(sigma, b, c)
            sigma, b, c = self._rom_to_sigma_b_c(rom, force_sigma_in_rhp)
            self._store_sigma_b_c(sigma, b, c)
            self._update_conv_data(sigma, rom, conv_crit)
            self._compute_conv_crit(rom, conv_crit, it)
            self._compute_error(rom, it, compute_errors)
            if self.conv_crit[-1] < tol:
                break

        return rom

    def reconstruct(self, u):
        """Reconstruct high-dimensional vector from reduced vector `u`."""
        raise TypeError(
            f'The reconstruct method is not available for {self.__class__.__name__}.'
        )


class GapIRKAReductor(GenericIRKAReductor):
    """Gap-IRKA reductor.

    Parameters
    ----------
    fom
        The full-order |LTIModel| to reduce.
    mu
        |Parameter|.
    """

    def __init__(self, fom, mu=None, solver_options=None):
        assert isinstance(fom, LTIModel)
        super().__init__(fom, mu=mu)
        self.solver_options = solver_options

    def reduce(self, rom0_params, tol=1e-4, maxit=100, num_prev=1, conv_crit='sigma',
               projection='orth'):
        r"""Reduce using gap-IRKA.

        See :cite:`BBG19` Algorithm 1.

        Parameters
        ----------
        rom0_params
            Can be:

            - order of the reduced model (a positive integer),
            - initial interpolation points (a 1D |NumPy array|),
            - dict with `'sigma'`, `'b'`, `'c'` as keys mapping to
              initial interpolation points (a 1D |NumPy array|), right
              tangential directions (|VectorArray| from
              `fom.input_space`), and left tangential directions
              (|VectorArray| from `fom.output_space`), all of the same
              length (the order of the reduced model),
            - initial reduced-order model (|LTIModel|).

            If the order of reduced model is given, initial
            interpolation data is generated randomly.
        tol
            Tolerance for the convergence criterion.
        maxit
            Maximum number of iterations.
        num_prev
            Number of previous iterations to compare the current
            iteration to. A larger number can avoid occasional cyclic
            behavior.
        conv_crit
            Convergence criterion:

            - `'sigma'`: relative change in interpolation points
            - `'htwogap'`: :math:`\mathcal{H}_2-gap` distance of
              reduced-order models divided by :math:`\mathcal{L}_2`
              norm of new reduced-order model
            - `'ltwo'`: relative :math:`\mathcal{L}_2` distance of
              reduced-order models
        projection
            Projection method:

            - `'orth'`: projection matrix is orthogonalized with respect
              to the Euclidean inner product,
            - `'biorth'`: projection matrix is orthogonalized with
              respect to the E product.

        Returns
        -------
        rom
            Reduced |LTIModel| model.
        """
        if not self.fom.cont_time:
            raise NotImplementedError

        self._clear_lists()
        sigma, b, c = self._rom0_params_to_sigma_b_c(rom0_params)
        self._store_sigma_b_c(sigma, b, c)
        assert projection in ('orth', 'biorth')

        self.logger.info('Starting gap IRKA')
        self._conv_data = (num_prev + 1) * [None]
        if conv_crit == 'sigma':
            self._conv_data[0] = sigma
        for it in range(maxit):
            self._pg_reductor = LTIBHIReductor(self.fom, mu=self.mu)
            rom = self._pg_reductor.reduce(sigma, b, c, projection)
            sigma, b, c, gap_rom = self._unstable_rom_to_sigma_b_c(rom)
            self._store_sigma_b_c(sigma, b, c)
            self._update_conv_data(sigma, gap_rom, rom, conv_crit)
            self._compute_conv_crit(rom, conv_crit, it)
            if self.conv_crit[-1] < tol:
                break

        return rom

    def _rom0_params_to_sigma_b_c(self, rom0_params):
        self.logger.info('Generating initial interpolation data')
        self._check_rom0_params(rom0_params)
        if isinstance(rom0_params, Integral):
            sigma, b, c = self._order_to_sigma_b_c(rom0_params)
        elif isinstance(rom0_params, np.ndarray):
            sigma = rom0_params
            _, b, c = self._order_to_sigma_b_c(len(rom0_params))
        elif isinstance(rom0_params, dict):
            sigma = rom0_params['sigma']
            b = rom0_params['b']
            c = rom0_params['c']
        else:
            sigma, b, c, _ = self._unstable_rom_to_sigma_b_c(rom0_params)
        return sigma, b, c

    def _unstable_rom_to_sigma_b_c(self, rom):
        poles, b, c, gap_rom = self._unstable_lti_to_poles_b_c(rom)
        return -poles, b, c, gap_rom

    def _unstable_lti_to_poles_b_c(self, rom):
        A = to_matrix(rom.A, format='dense')
        B = to_matrix(rom.B, format='dense')
        C = to_matrix(rom.C, format='dense')

        options = self.solver_options

        if isinstance(rom.E, IdentityOperator):
            P = solve_ricc_dense(A, None, B, C, options=options)
            F = P @ C.T
            AF = A - F @ C
            poles, X = spla.eig(AF)
            EX = X
        else:
            E = to_matrix(rom.E, format='dense')
            P = solve_ricc_dense(A, E, B, C, options=options)
            F = E @ P @ C.T
            AF = A - F @ C
            poles, X = spla.eig(AF, E)
            EX = E @ X
        b = spla.solve(EX, B)
        c = (C @ X).T
        mFB = np.concatenate((-F, B), axis=1)
        gap_rom = LTIModel.from_matrices(AF, mFB, C, E=None if isinstance(rom.E, IdentityOperator) else E)
        return poles, b, c, gap_rom

    def _update_conv_data(self, sigma, gap_rom, rom, conv_crit):
        del self._conv_data[-1]
        if conv_crit == 'sigma':
            self._conv_data.insert(0, sigma)
        elif conv_crit == 'htwogap':
            self._conv_data.insert(0, gap_rom)
        elif conv_crit == 'ltwo':
            self._conv_data.insert(0, rom)

    def _compute_conv_crit(self, rom, conv_crit, it):
        if conv_crit == 'sigma':
            sigma = self._conv_data[0]
            dist = min(spla.norm((sigma_old - sigma) / sigma_old, ord=np.inf)
                       for sigma_old in self._conv_data[1:]
                       if sigma_old is not None)
        elif conv_crit == 'htwogap':
            gap_rom = self._conv_data[0]
            A = to_matrix(rom.A, format='dense')
            C = to_matrix(rom.C, format='dense')
            F = -to_matrix(gap_rom.B, format='dense')[:, :C.shape[0]]
            D = np.eye(rom.C.range.dim)
            E = to_matrix(rom.E, format='dense') if isinstance(rom.E, IdentityOperator) else None

            mi = LTIModel.from_matrices(A, F, C, D, E)
            dist = min((gap_rom_old - gap_rom).h2_norm() * mi.linf_norm() / rom.l2_norm()
                       if gap_rom_old is not None
                       else np.inf
                       for gap_rom_old in self._conv_data[1:])
        elif conv_crit == 'ltwo':
            dist = min((rom_old - rom).l2_norm() / rom.l2_norm()
                       if rom_old is not None
                       else np.inf
                       for rom_old in self._conv_data[1:])

        self.conv_crit.append(dist)
        self.logger.info(f'Convergence criterion in iteration {it + 1}: {dist:e}')



#!/usr/bin/env python
# This file is part of the pyMOR project (http://www.pymor.org).
# Copyright 2013-2020 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (http://opensource.org/licenses/BSD-2-Clause)

import numpy as np
import scipy.sparse as sps
from typer import Argument, run


from pymor.core.config import config
from pymor.core.logger import set_log_levels
from pymor.models.iosys import LTIModel
from pymor.reductors.bt import FDBTReductor
from pymor.reductors.h2 import GapIRKAReductor


def run_mor_method(lti, reductor, reductor_short_name, r, **reduce_kwargs):
    """Run a model order reduction method.

    Parameters
    ----------
    lti
        The full-order |LTIModel|.
    w
        Array of frequencies.
    reductor
        The reductor object.
    reductor_short_name
        A short name for the reductor.
    r
        The order of the reduced-order model.
    reduce_kwargs
        Optional keyword arguments for the reduce method.
    """
    # Reduction
    rom = reductor.reduce(r, **reduce_kwargs)
    err = lti - rom

    # Errors
    if config.HAVE_SLYCOT:
        print(f'{reductor_short_name} relative L_inf-error:  {err.linf_norm() / lti.linf_norm():e}')
    else:
        print('Skipped L_inf-norm calculation due to missing slycot.')


def main(
        l: float = Argument(50, help='Parameter for instability.'),
        r: int = Argument(10, help='Order of the ROMs.'),
):
    r"""1D unstable heat equation demo.

    Discretization of the PDE:

    .. math::

        \begin{align}
            \partial_t T(\xi, t) &= \partial_{\xi \xi} T(\xi, t) + \lambda T(\xi, t),
            & 0 < \xi < 1,\ t > 0, \\
            -\partial_\xi T(0, t) & = -T(0, t) + u(t),
            & t > 0, \\
            \partial_\xi T(1, t) & = -T(1, t),
            & t > 0, \\
            y(t) & = T(1, t),
            & t > 0
        \end{align}

    """
    set_log_levels({'pymor.algorithms.gram_schmidt.gram_schmidt': 'WARNING'})

    k = 50
    n = 2 * k + 1

    E = sps.eye(n, format='lil')
    E[0, 0] = E[-1, -1] = 0.5
    E = E.tocsc()

    d0 = n * [-2 * (n - 1)**2 + l]
    d1 = (n - 1) * [(n - 1)**2]
    A = sps.diags([d1, d0, d1], [-1, 0, 1], format='lil')
    A[0, 0] = A[-1, -1] = -n * (n - 1) + l / 2
    A = A.tocsc()

    B = np.zeros((n, 1))
    B[0, 0] = n - 1

    C = np.zeros((1, n))
    C[0, -1] = 1

    lti = LTIModel.from_matrices(A, B, C, E=E)

    ast_spectrum = lti.get_ast_spectrum()
    print(f'Anti-stable system poles:  {" ".join(str(x) for x in ast_spectrum[1])}')

    # Norms of the system
    print(f'FOM L_2-norm:    {lti.l2_norm():e}')
    if config.HAVE_SLYCOT:
        print(f'FOM L_inf-norm:  {lti.hinf_norm():e}')
    else:
        print('Skipped L_inf-norm calculation due to missing slycot.')

    # Model order reduction
    run_mor_method(lti, FDBTReductor(lti), 'FDBT', r, tol=1e-5)
    run_mor_method(lti, GapIRKAReductor(lti), 'GapIRKA', r, tol=1e-5)


if __name__ == '__main__':
    run(main)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import pytest
import numpy as np
from hypothesis import given, assume, settings
from hypothesis import strategies as hyst

import pymor
from pymor.algorithms.basic import almost_equal, project_array, relative_error
from pymor.algorithms.gram_schmidt import gram_schmidt
if pymor.config.HAVE_NGSOLVE:
    from pymor.bindings.ngsolve import NGSolveVectorSpace
else:
    class NGSolveVectorSpace:
        pass
from pymor.operators.numpy import NumpyMatrixOperator
from pymor.vectorarrays.numpy import NumpyVectorSpace
from pymortests.strategies import valid_inds, valid_inds_of_same_length
from pymortests.vectorarray import indexed
import pymortests.strategies as pyst


@pyst.given_vector_arrays(
    count=2,
    tolerances=hyst.sampled_from([(1e-5, 1e-8), (1e-10, 1e-12), (0., 1e-8), (1e-5, 1e-8)]),
    sup_norm=hyst.booleans(),
)
def test_almost_equal(vector_arrays, tolerances, sup_norm):
    v1, v2 = vector_arrays
    rtol, atol = tolerances
    try:
        dv1 = v1.to_numpy()
        dv2 = v2.to_numpy()
    except NotImplementedError:
        dv1 = dv2 = None
    for ind1, ind2 in valid_inds_of_same_length(v1, v2):
        r = almost_equal(v1[ind1], v2[ind2], sup_norm=sup_norm, rtol=rtol, atol=atol)
        assert isinstance(r, np.ndarray)
        assert r.shape == (v1.len_ind(ind1),)
        if dv1 is not None:
            if dv2.shape[1] == 0:
                continue
            assert np.all(r == (np.linalg.norm(indexed(dv1, ind1) - indexed(dv2, ind2),
                                               ord=np.inf if sup_norm else None,
                                               axis=1)
                                <= atol + rtol * np.linalg.norm(indexed(dv2, ind2),
                                                                ord=np.inf if sup_norm else None,
                                                                axis=1)))


def test_almost_equal_product(operator_with_arrays_and_products):
    _, _, v1, _, product, _ = operator_with_arrays_and_products
    if len(v1) < 2:
        return
    v2 = v1.empty()
    v2.append(v1[:len(v1) // 2])
    for ind1, ind2 in valid_inds_of_same_length(v1, v2):
        for rtol, atol in ((1e-5, 1e-8), (1e-10, 1e-12), (0., 1e-8), (1e-5, 1e-8)):

            r = almost_equal(v1[ind1], v2[ind2], product=product)
            assert isinstance(r, np.ndarray)
            assert r.shape == (v1.len_ind(ind1),)
            assert np.all(r == ((v1[ind1] - v2[ind2]).norm(product)
                                <= atol + rtol * (v2[ind2]).norm(product)))


@pyst.given_vector_arrays(
    count=1,
    index_strategy=pyst.pairs_same_length,
    tolerances=hyst.sampled_from([(1e-5, 1e-8), (1e-10, 1e-12), (0., 1e-8), (1e-5, 1e-8), (1e-12, 0.)]),
    sup_norm=hyst.booleans(),
)
@settings(print_blob=True)
def test_almost_equal_self(vectors_and_indices, tolerances, sup_norm):
    v, (ind, _) = vectors_and_indices
    rtol, atol = tolerances
    r = almost_equal(v[ind], v[ind], sup_norm=sup_norm)
    assert isinstance(r, np.ndarray)
    assert r.shape == (v.len_ind(ind),)
    assert np.all(r)
    # TODO drop assumptions
    # the first assumption here is a direct translation of the old loop abort
    assume(v.len_ind(ind) > 0 and np.max(v[ind].sup_norm() > 0))
    # the second one accounts for old input missing very-near zero data
    if sup_norm:
        norm = lambda U: U.sup_norm()
    else:
        norm = lambda U: U.norm()
    tol_min = np.min(np.abs(tolerances))
    v_n_min = np.min(norm(v[ind]))
    assume(v_n_min > tol_min)

    c = v.copy()
    c.scal(atol * (1 - 1e-10) / (np.max(norm(v[ind]))))
    assert np.all(almost_equal(c[ind], c.zeros(v.len_ind(ind)), atol=atol, rtol=rtol, sup_norm=sup_norm))

    if atol > 0:
        c = v.copy()
        c.scal(2. * atol / (np.max(norm(v[ind]))))
        assert not np.all(almost_equal(c[ind], c.zeros(v.len_ind(ind)), atol=atol, rtol=rtol, sup_norm=sup_norm))

    c = v.copy()
    c.scal(1. + rtol * 0.9)
    assert np.all(almost_equal(c[ind], v[ind], atol=atol, rtol=rtol, sup_norm=sup_norm))

    if rtol > 0:
        c = v.copy()
        c.scal(2. + rtol * 1.1)
        assert not np.all(almost_equal(c[ind], v[ind], atol=atol, rtol=rtol, sup_norm=sup_norm))

    c = v.copy()
    c.scal(1. + atol * 0.9 / np.max(norm(v[ind])))
    assert np.all(almost_equal(c[ind], v[ind], atol=atol, rtol=rtol, sup_norm=sup_norm))

    if atol > 0 or rtol > 0:
        c = v.copy()
        c.scal(1 + rtol * 1.1 + atol * 1.1 / np.max(norm(v[ind])))
        assert not np.all(almost_equal(c[ind], v[ind], atol=atol, rtol=rtol, sup_norm=sup_norm))


def test_almost_equal_self_product(operator_with_arrays_and_products):
    _, _, v, _, product, _ = operator_with_arrays_and_products
    for ind in valid_inds(v):
        for rtol, atol in ((1e-5, 1e-8), (1e-10, 1e-12), (0., 1e-8), (1e-5, 1e-8), (1e-12, 0.)):
            r = almost_equal(v[ind], v[ind], product=product)
            assert isinstance(r, np.ndarray)
            assert r.shape == (v.len_ind(ind),)
            assert np.all(r)
            if v.len_ind(ind) == 0 or np.max(v[ind].sup_norm() == 0):
                continue

            c = v.copy()
            c.scal(atol * (1 - 1e-10) / (np.max((v[ind]).norm(product))))
            assert np.all(almost_equal(c[ind], c.zeros(v.len_ind(ind)), atol=atol, rtol=rtol, product=product))

            if atol > 0:
                c = v.copy()
                c.scal(2. * atol / (np.max((v[ind]).norm(product))))
                assert not np.all(almost_equal(c[ind], c.zeros(v.len_ind(ind)), atol=atol, rtol=rtol, product=product))

            c = v.copy()
            c.scal(1. + rtol * 0.9)
            assert np.all(almost_equal(c[ind], v[ind], atol=atol, rtol=rtol, product=product))

            if rtol > 0:
                c = v.copy()
                c.scal(2. + rtol * 1.1)
                assert not np.all(almost_equal(c[ind], v[ind], atol=atol, rtol=rtol, product=product))

            c = v.copy()
            c.scal(1. + atol * 0.9 / np.max(np.max((v[ind]).norm(product))))
            assert np.all(almost_equal(c[ind], v[ind], atol=atol, rtol=rtol, product=product))

            if atol > 0 or rtol > 0:
                c = v.copy()
                c.scal(1 + rtol * 1.1 + atol * 1.1 / np.max(np.max((v[ind]).norm(product))))
                assert not np.all(almost_equal(c[ind], v[ind], atol=atol, rtol=rtol, product=product))


@pyst.given_vector_arrays(count=2, compatible=False)
def test_almost_equal_incompatible(vector_arrays):
    v1, v2 = vector_arrays
    for ind1, ind2 in valid_inds_of_same_length(v1, v2):
        for sup_norm in (False, True):
            c1, c2 = v1.copy(), v2.copy()
            with pytest.raises(Exception):
                almost_equal(c1[ind1], c2[ind2], sup_norm=sup_norm)


@given(pyst.base_vector_arrays(count=2))
@settings(deadline=None)
def test_project_array(arrays):
    U, basis = arrays
    U_p = project_array(U, basis, orthonormal=False)
    onb = gram_schmidt(basis)
    U_p2 = project_array(U, onb, orthonormal=True)
    err = relative_error(U_p, U_p2)
    tol = np.finfo(np.float64).eps * np.linalg.cond(basis.gramian()) * 100.
    assert np.all(err < tol)


def test_project_array_with_product():
    np.random.seed(123)
    U = NumpyVectorSpace.from_numpy(np.random.random((1, 10)))
    basis = NumpyVectorSpace.from_numpy(np.random.random((3, 10)))
    product = np.random.random((10, 10))
    product = NumpyMatrixOperator(product.T.dot(product))
    U_p = project_array(U, basis, product=product, orthonormal=False)
    onb = gram_schmidt(basis, product=product)
    U_p2 = project_array(U, onb, product=product, orthonormal=True)
    tol = 3e-10
    assert np.all(relative_error(U_p, U_p2, product) < tol)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import os
import platform
import sys


if sys.version_info.major < 3:
    raise RuntimeError('pyMOR requires Python 3.6 or newer')
elif sys.version_info.major == 3:
    if sys.version_info.minor < 6:
        raise RuntimeError('pyMOR requires Python 3.6 or newer.')
    if sys.version_info.minor == 6 and platform.python_implementation() != 'CPython':
        # dicts are only guaranteed to be insertion ordered for Python 3.7 or newer.
        # For CPython this is already the case for Python 3.6.
        raise RuntimeError('pyMOR requires Python 3.7 or newer for non-CPython interpreters.')


def _init_mpi():
    """provides a way to manually set the thread init mode for MPI if necessary.
    Needs to happen as early as possible, otherwise mpi4py might auto-init somewhere else.


    """
    try:
        import mpi4py
    except ImportError:
        return

    finalize = os.environ.get('PYMOR_MPI_FINALIZE', mpi4py.rc.finalize if mpi4py.rc.finalize is not None else False)
    mpi4py.rc(initialize=False, finalize=finalize)
    from mpi4py import MPI
    if not MPI.Is_initialized():
        required_level = int(os.environ.get('PYMOR_MPI_INIT_THREAD', MPI.THREAD_MULTIPLE))
        supported_lvl = MPI.Init_thread(required_level)
        if supported_lvl < required_level:
            print(f'MPI does support threading level {required_level}, running with {supported_lvl} instead', flush=True)
    try:
        # this solves sporadic mpi calls happening after finalize
        import petsc4py
        petsc4py.init()
    except ImportError:
        return


_init_mpi()

from pymor.core.config import config
from pymor.core.defaults import load_defaults_from_file


import pymor.version as _version
revstring = _version.get_versions()['version']
__version__ = str(revstring)

if 'PYMOR_DEFAULTS' in os.environ:
    filename = os.environ['PYMOR_DEFAULTS']
    if filename in ('', 'NONE'):
        print('Not loading any pyMOR defaults from config file')
    else:
        for fn in filename.split(':'):
            if not os.path.exists(fn):
                raise IOError('Cannot load pyMOR defaults from file ' + fn)
            print('Loading pyMOR defaults from file ' + fn + ' (set by PYMOR_DEFAULTS)')
            load_defaults_from_file(fn)
else:
    filename = os.path.join(os.getcwd(), 'pymor_defaults.py')
    if os.path.exists(filename):
        from pymor.tools.io import file_owned_by_current_user
        if not file_owned_by_current_user(filename):
            raise IOError('Cannot load pyMOR defaults from config file ' + filename
                          + ': not owned by user running Python interpreter')
        print('Loading pyMOR defaults from file ' + filename)
        load_defaults_from_file(filename)

from pymor.core.logger import set_log_levels, set_log_format
set_log_levels()
set_log_format()

from pymor.tools import mpi
if mpi.parallel and mpi.event_loop_settings()['auto_launch']:
    if mpi.rank0:
        import atexit
        @atexit.register
        def quit_event_loop():
            if not mpi.finished:
                mpi.quit()
    else:
        print(f'Rank {mpi.rank}: MPI parallel run detected. Launching event loop ...')
    mpi.launch_event_loop()
    if not mpi.rank0:
        import sys
        sys.exit(0)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import time
import numpy as np
import functools

from pymor.core.logger import getLogger


class Timer:
    """Class for finding code runtime.

    You can use me as a context manager, plain instance or decorator to time execution
    of a code scope::

        with Timer() as timer:
            do_some_stuff()
            do more stuff()
        #outputs time in (s)

        ### OR ###

        @timing.Timer('name', logging.debug)
        def function(*args):
            do_stuff

        function(1)
        #outputs time in (s) to logging.debug

        ### OR ###

        timer = timing.Timer()
        timer.start()
        do_stuff()
        timer.stop()
        print(timer.dt)
    """

    def __init__(self, section, log=getLogger(__name__)):
        self._section = section
        self._log = log
        self._start = 0

    def start(self):
        self.dt = -1
        self._start = time.perf_counter()

    def stop(self):
        self.dt = time.perf_counter() - self._start

    def __enter__(self):
        self.start()

    def __exit__(self, type_, value, traceback):
        self.stop()
        self._log.info('Execution of %s took %f (s)', self._section, self.dt)

    def __call__(self, func):
        func.decorated = self

        @functools.wraps(func)
        def new_func(*args, **kwargs):
            with self:
                return func(*args, **kwargs)
        return new_func


def busywait(amount):
    arr = np.arange(1000)
    for _ in range(amount):
        np.random.shuffle(arr)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np
from hypothesis import settings, assume, given

from pymor.algorithms.basic import almost_equal
from pymor.algorithms.gram_schmidt import gram_schmidt, gram_schmidt_biorth
from pymor.core.logger import log_levels
from pymor.algorithms.basic import contains_zero_vector
from pymortests.base import runmodule
import pymortests.strategies as pyst


@pyst.given_vector_arrays()
@settings(deadline=20000)
def test_gram_schmidt(vector_array):
    U = vector_array
    # TODO assumption here masks a potential issue with the algorithm
    #      where it fails in del instead of a proper error
    assume(len(U) > 1 or not contains_zero_vector(U))

    V = U.copy()
    onb = gram_schmidt(U, copy=True)
    assert np.all(almost_equal(U, V))
    assert np.allclose(onb.inner(onb), np.eye(len(onb)))
    # TODO maybe raise tolerances again
    assert np.all(almost_equal(U, onb.lincomb(onb.inner(U).T), atol=1e-13, rtol=1e-13))

    onb2 = gram_schmidt(U, copy=False)
    assert np.all(almost_equal(onb, onb2))
    assert np.all(almost_equal(onb, U))


@pyst.given_vector_arrays()
@settings(deadline=None)
def test_gram_schmidt_with_R(vector_array):
    U = vector_array
    # TODO assumption here masks a potential issue with the algorithm
    #      where it fails in del instead of a proper error
    assume(len(U) > 1 or not contains_zero_vector(U))

    V = U.copy()
    onb, R = gram_schmidt(U, return_R=True, copy=True)
    assert np.all(almost_equal(U, V))
    assert np.allclose(onb.inner(onb), np.eye(len(onb)))
    lc = onb.lincomb(onb.inner(U).T)
    rtol = atol = 1e-13
    assert np.all(almost_equal(U, lc, rtol=rtol, atol=atol))
    assert np.all(almost_equal(V, onb.lincomb(R.T), rtol=rtol, atol=atol))

    onb2, R2 = gram_schmidt(U, return_R=True, copy=False)
    assert np.all(almost_equal(onb, onb2))
    assert np.all(R == R2)
    assert np.all(almost_equal(onb, U))


def test_gram_schmidt_with_product(operator_with_arrays_and_products):
    _, _, U, _, p, _ = operator_with_arrays_and_products

    V = U.copy()
    onb = gram_schmidt(U, product=p, copy=True)
    assert np.all(almost_equal(U, V))
    assert np.allclose(p.apply2(onb, onb), np.eye(len(onb)))
    assert np.all(almost_equal(U, onb.lincomb(p.apply2(onb, U).T), rtol=1e-13))

    onb2 = gram_schmidt(U, product=p, copy=False)
    assert np.all(almost_equal(onb, onb2))
    assert np.all(almost_equal(onb, U))


def test_gram_schmidt_with_product_and_R(operator_with_arrays_and_products):
    _, _, U, _, p, _ = operator_with_arrays_and_products

    V = U.copy()
    onb, R = gram_schmidt(U, product=p, return_R=True, copy=True)
    assert np.all(almost_equal(U, V))
    assert np.allclose(p.apply2(onb, onb), np.eye(len(onb)))
    assert np.all(almost_equal(U, onb.lincomb(p.apply2(onb, U).T), rtol=1e-13))
    assert np.all(almost_equal(U, onb.lincomb(R.T)))

    onb2, R2 = gram_schmidt(U, product=p, return_R=True, copy=False)
    assert np.all(almost_equal(onb, onb2))
    assert np.all(R == R2)
    assert np.all(almost_equal(onb, U))


@given(pyst.base_vector_arrays(count=2))
@settings(deadline=None)
def test_gram_schmidt_biorth(vector_arrays):
    U1, U2 = vector_arrays

    V1 = U1.copy()
    V2 = U2.copy()

    # this is the default used in gram_schmidt_biorth
    check_tol = 1e-3
    with log_levels({'pymor.algorithms.gram_schmidt.gram_schmidt_biorth': 'ERROR'}):
        A1, A2 = gram_schmidt_biorth(U1, U2, copy=True, check_tol=check_tol)
    assert np.all(almost_equal(U1, V1))
    assert np.all(almost_equal(U2, V2))
    assert np.allclose(A2.inner(A1), np.eye(len(A1)), atol=check_tol)
    c = np.linalg.cond(A1.to_numpy()) * np.linalg.cond(A2.to_numpy())
    assert np.all(almost_equal(U1, A1.lincomb(A2.inner(U1).T), rtol=c * 1e-14))
    assert np.all(almost_equal(U2, A2.lincomb(A1.inner(U2).T), rtol=c * 1e-14))

    with log_levels({'pymor.algorithms.gram_schmidt.gram_schmidt_biorth': 'ERROR'}):
        B1, B2 = gram_schmidt_biorth(U1, U2, copy=False)
    assert np.all(almost_equal(A1, B1))
    assert np.all(almost_equal(A2, B2))
    assert np.all(almost_equal(A1, U1))
    assert np.all(almost_equal(A2, U2))


def test_gram_schmidt_biorth_with_product(operator_with_arrays_and_products):
    _, _, U, _, p, _ = operator_with_arrays_and_products
    if U.dim < 2:
        return
    l = len(U) // 2
    l = min((l, U.dim - 1))
    if l < 1:
        return
    U1 = U[:l].copy()
    U2 = U[l:2 * l].copy()

    V1 = U1.copy()
    V2 = U2.copy()
    A1, A2 = gram_schmidt_biorth(U1, U2, product=p, copy=True)
    assert np.all(almost_equal(U1, V1))
    assert np.all(almost_equal(U2, V2))
    assert np.allclose(p.apply2(A2, A1), np.eye(len(A1)))
    c = np.linalg.cond(A1.to_numpy()) * np.linalg.cond(p.apply(A2).to_numpy())
    assert np.all(almost_equal(U1, A1.lincomb(p.apply2(A2, U1).T), rtol=c * 1e-14))
    assert np.all(almost_equal(U2, A2.lincomb(p.apply2(A1, U2).T), rtol=c * 1e-14))

    B1, B2 = gram_schmidt_biorth(U1, U2, product=p, copy=False)
    assert np.all(almost_equal(A1, B1))
    assert np.all(almost_equal(A2, B2))
    assert np.all(almost_equal(A1, U1))
    assert np.all(almost_equal(A2, U2))


if __name__ == "__main__":
    runmodule(filename=__file__)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

"""This module contains methods for object serialization.

Instead of importing serialization functions from Python's
:mod:`pickle` module directly, you should use the `dump`, `dumps`,
`load`, `loads` functions defined here. In particular, these
methods will use :func:`dumps_function` to serialize
function objects which cannot be pickled by Python's standard
methods. Note, however, pickling such methods should be avoided
since the implementation of :func:`dumps_function` uses non-portable
implementation details of CPython to achieve its goals.
"""

import marshal
import opcode
from types import CodeType, FunctionType, ModuleType
import pickle
from io import BytesIO as IOtype
import platform


PicklingError = pickle.PicklingError
UnpicklingError = pickle.UnpicklingError
PROTOCOL = pickle.HIGHEST_PROTOCOL


# on CPython provide pickling methods which use
# dumps_function in case pickling of a function fails
if platform.python_implementation() == 'CPython':

    def dump(obj, file, protocol=None):
        pickler = pickle.Pickler(file, protocol=PROTOCOL)
        pickler.persistent_id = _function_pickling_handler
        pickler.dump(obj)

    def dumps(obj, protocol=None):
        file = IOtype()
        pickler = pickle.Pickler(file, protocol=PROTOCOL)
        pickler.persistent_id = _function_pickling_handler
        pickler.dump(obj)
        return file.getvalue()

    def load(file):
        unpickler = pickle.Unpickler(file)
        unpickler.persistent_load = _function_unpickling_handler
        return unpickler.load()

    def loads(str):
        file = IOtype(str)
        unpickler = pickle.Unpickler(file)
        unpickler.persistent_load = _function_unpickling_handler
        return unpickler.load()

else:
    from functools import partial
    dump = partial(pickle.dump, protocol=PROTOCOL)
    dumps = partial(pickle.dumps, protocol=PROTOCOL)
    load = pickle.load
    loads = pickle.loads


def _generate_opcode(code_object):
    import dis
    for ins in dis.get_instructions(code_object):
        yield (ins.opcode, ins.arg)


def _global_names(code_object):
    """Return all names in code_object.co_names which are used in a LOAD_GLOBAL statement."""
    LOAD_GLOBAL = opcode.opmap['LOAD_GLOBAL']
    indices = {i for o, i in _generate_opcode(code_object) if o == LOAD_GLOBAL}
    names = code_object.co_names
    result = {names[i] for i in indices}

    # On Python 3, comprehensions have their own scope. This is implemented
    # by generating a new code object for the comprehension which is stored
    # as a constant of the enclosing function's code object. If the comprehension
    # refers to global names, these names are listed in co_names of the code
    # object for the comprehension, so we have to look at these code objects as
    # well:
    for const in code_object.co_consts:
        if type(const) is CodeType:
            result.update(_global_names(const))

    return result


class Module:

    def __init__(self, mod):
        self.mod = mod

    def __getstate__(self):
        if not hasattr(self.mod, '__package__'):
            raise PicklingError
        return self.mod.__package__

    def __setstate__(self, s):
        self.mod = __import__(s)


def dumps_function(function):
    """Tries hard to pickle a function object:

        1. The function's code object is serialized using the :mod:`marshal` module.
        2. For all global names used in the function's code object the corresponding
           object in the function's global namespace is pickled. In case this object
           is a module, the modules __package__ name is pickled.
        3. All default arguments are pickled.
        4. All objects in the function's closure are pickled.

    Note that also this is heavily implementation specific and will probably only
    work with CPython. If possible, avoid using this method.
    """
    closure = None if function.__closure__ is None else [c.cell_contents for c in function.__closure__]
    code = marshal.dumps(function.__code__)
    func_globals = function.__globals__

    def wrap_modules(x):
        return Module(x) if isinstance(x, ModuleType) else x

    # note that global names in function.func_code can also refer to builtins ...
    globals_ = {k: wrap_modules(func_globals[k]) for k in _global_names(function.__code__) if k in func_globals}

    return dumps((function.__name__, code, globals_, function.__defaults__, closure, function.__dict__,
                  function.__doc__, function.__qualname__, function.__kwdefaults__, function.__annotations__))


def loads_function(s):
    """Restores a function serialized with :func:`dumps_function`."""
    name, code, globals_, defaults, closure, func_dict, doc, qualname, kwdefaults, annotations = loads(s)
    code = marshal.loads(code)
    for k, v in globals_.items():
        if isinstance(v, Module):
            globals_[k] = v.mod
    if closure is not None:
        import ctypes
        ctypes.pythonapi.PyCell_New.restype = ctypes.py_object
        ctypes.pythonapi.PyCell_New.argtypes = [ctypes.py_object]
        closure = tuple(ctypes.pythonapi.PyCell_New(c) for c in closure)
    globals_['__builtins__'] = __builtins__
    r = FunctionType(code, globals_, name, defaults, closure)
    r.__dict__ = func_dict
    r.__doc__ = doc
    r.__qualname__ = qualname
    r.__kwdefaults__ = kwdefaults
    r.__annotations__ = annotations
    return r


def _function_pickling_handler(f):
    if f.__class__ is FunctionType:
        if f.__module__ != '__main__':
            try:
                return b'A' + pickle.dumps(f)
            except (AttributeError, TypeError, PicklingError):
                return b'B' + dumps_function(f)
        else:
            return b'B' + dumps_function(f)
    else:
        return None


def _function_unpickling_handler(persid):
    mode, data = persid[0], persid[1:]
    if mode == b'A'[0]:
        return pickle.loads(data)
    elif mode == b'B'[0]:
        return loads_function(data)
    else:
        raise UnpicklingError



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import pytest

from pymor.analyticalproblems.burgers import burgers_problem, burgers_problem_2d
from pymor.analyticalproblems.domaindescriptions import RectDomain
from pymor.analyticalproblems.elliptic import StationaryProblem
from pymor.analyticalproblems.functions import GenericFunction, ConstantFunction, LincombFunction
from pymor.analyticalproblems.helmholtz import helmholtz_problem
from pymor.analyticalproblems.thermalblock import thermal_block_problem
from pymor.parameters.functionals import ExpressionParameterFunctional


picklable_thermalblock_problems = [
    thermal_block_problem(),
    thermal_block_problem(num_blocks=(3, 2)),
    thermal_block_problem(num_blocks=(1, 1)),
    thermal_block_problem(num_blocks=(2, 2), parameter_range=(1., 100.)),
]


non_picklable_thermalblock_problems = [
    thermal_block_problem(
        num_blocks=(1, 3),
        parameter_range=(0.4, 0.5)).with_(
            rhs=GenericFunction(dim_domain=2, mapping=lambda X: X[..., 0] + X[..., 1])),
]


thermalblock_problems = picklable_thermalblock_problems + non_picklable_thermalblock_problems


burgers_problems = [
    burgers_problem(),
    burgers_problem(v=0.2, circle=False),
    burgers_problem(v=0.4, initial_data_type='bump'),
    burgers_problem(parameter_range=(1., 1.3)),
    burgers_problem_2d(),
    burgers_problem_2d(torus=False, initial_data_type='bump', parameter_range=(1.3, 1.5)),
]


picklable_elliptic_problems = [
    StationaryProblem(
        domain=RectDomain(),
        rhs=ConstantFunction(dim_domain=2, value=1.),
    ),
    helmholtz_problem(),
]


non_picklable_elliptic_problems = [
    StationaryProblem(
        domain=RectDomain(),
        rhs=ConstantFunction(dim_domain=2, value=21.),
        diffusion=LincombFunction(
            [GenericFunction(dim_domain=2, mapping=lambda X, p=p: X[..., 0]**p)
             for p in range(5)],
            [ExpressionParameterFunctional(f'max(exp * [1,0] + [0, {m}])', parameters={'exp': 1})
             for m in range(5)],
        ),
    ),
]


elliptic_problems = picklable_thermalblock_problems + non_picklable_elliptic_problems


@pytest.fixture(params=elliptic_problems + thermalblock_problems + burgers_problems)
def analytical_problem(request):
    return request.param


@pytest.fixture(params=picklable_elliptic_problems + picklable_thermalblock_problems + burgers_problems)
def picklable_analytical_problem(request):
    return request.param



#!/usr/bin/env python
# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)


import numpy as np
from typer import Argument, run

from pymor.basic import *


def main(
    grid_intervals: int = Argument(..., help='Grid interval count.'),
    training_samples: int = Argument(..., help='Number of samples used for training the reduced basis.')
):
    """Example script for solving linear PDE-constrained parameter optimization problems"""
    fom, mu_bar = create_fom(grid_intervals)

    parameter_space = fom.parameters.space(0, np.pi)
    ranges = parameter_space.ranges['diffusion']

    initial_guess = fom.parameters.parse([0.25, 0.5])

    def fom_objective_functional(mu):
        return fom.output(mu)

    def fom_gradient_of_functional(mu):
        return fom.output_d_mu(fom.parameters.parse(mu), return_array=True, use_adjoint=True)

    from functools import partial
    from scipy.optimize import minimize
    from time import perf_counter

    opt_fom_minimization_data = {'num_evals': 0,
                                 'evaluations': [],
                                 'evaluation_points': [],
                                 'time': np.inf}
    tic = perf_counter()
    opt_fom_result = minimize(partial(record_results, fom_objective_functional,
                                      fom.parameters.parse, opt_fom_minimization_data),
                              initial_guess.to_numpy(),
                              method='L-BFGS-B',
                              jac=fom_gradient_of_functional,
                              bounds=(ranges, ranges),
                              options={'ftol': 1e-15})
    opt_fom_minimization_data['time'] = perf_counter()-tic

    reference_mu = opt_fom_result.x

    from pymor.algorithms.greedy import rb_greedy
    from pymor.reductors.coercive import CoerciveRBReductor
    from pymor.parameters.functionals import MinThetaParameterFunctional

    coercivity_estimator = MinThetaParameterFunctional(fom.operator.coefficients, mu_bar)

    training_set = parameter_space.sample_uniformly(training_samples)

    RB_reductor = CoerciveRBReductor(fom, product=fom.energy_product, coercivity_estimator=coercivity_estimator)
    RB_greedy_data = rb_greedy(fom, RB_reductor, training_set, atol=1e-2)
    rom = RB_greedy_data['rom']

    def rom_objective_functional(mu):
        return rom.output(mu)

    def rom_gradient_of_functional(mu):
        return rom.output_d_mu(fom.parameters.parse(mu), return_array=True, use_adjoint=True)

    opt_rom_minimization_data = {'num_evals': 0,
                                 'evaluations': [],
                                 'evaluation_points': [],
                                 'time': np.inf,
                                 'offline_time': RB_greedy_data['time']}

    tic = perf_counter()
    opt_rom_result = minimize(partial(record_results, rom_objective_functional, fom.parameters.parse,
                                      opt_rom_minimization_data),
                              initial_guess.to_numpy(),
                              method='L-BFGS-B',
                              jac=rom_gradient_of_functional,
                              bounds=(ranges, ranges),
                              options={'ftol': 1e-15})
    opt_rom_minimization_data['time'] = perf_counter()-tic

    print("\nResult of optimization with FOM model and adjoint gradient")
    report(opt_fom_result, fom.parameters.parse, opt_fom_minimization_data, reference_mu)
    print("Result of optimization with ROM model and adjoint gradient")
    report(opt_rom_result, fom.parameters.parse, opt_rom_minimization_data, reference_mu)


def create_fom(grid_intervals, vector_valued_output=False):
    domain = RectDomain(([-1, -1], [1, 1]))
    indicator_domain = ExpressionFunction(
        '(-2/3. <= x[0]) * (x[0] <= -1/3.) * (-2/3. <= x[1]) * (x[1] <= -1/3.) * 1. \
       + (-2/3. <= x[0]) * (x[0] <= -1/3.) *  (1/3. <= x[1]) * (x[1] <=  2/3.) * 1.',
        dim_domain=2)
    rest_of_domain = ConstantFunction(1, 2) - indicator_domain

    f = ExpressionFunction('0.5*pi*pi*cos(0.5*pi*x[0])*cos(0.5*pi*x[1])', dim_domain=2)

    parameters = {'diffusion': 2}
    thetas = [
        ExpressionParameterFunctional(
            '1.1 + sin(diffusion[0])*diffusion[1]',
            parameters,
            derivative_expressions={'diffusion': ['cos(diffusion[0])*diffusion[1]',
                                                  'sin(diffusion[0])']},
        ),
        ExpressionParameterFunctional(
            '1.1 + sin(diffusion[1])',
            parameters,
            derivative_expressions={'diffusion': ['0', 'cos(diffusion[1])']},
        ),
    ]
    diffusion = LincombFunction([rest_of_domain, indicator_domain], thetas)

    theta_J = ExpressionParameterFunctional('1 + 1/5 * diffusion[0] + 1/5 * diffusion[1]', parameters,
                                            derivative_expressions={'diffusion': ['1/5', '1/5']})

    if vector_valued_output:
        problem = StationaryProblem(domain, f, diffusion, outputs=[('l2', f * theta_J), ('l2', f * 0.5 * theta_J)])
    else:
        problem = StationaryProblem(domain, f, diffusion, outputs=[('l2', f * theta_J)])

    print('Discretize ...')
    mu_bar = problem.parameters.parse([np.pi/2, np.pi/2])
    fom, _ = discretize_stationary_cg(problem, diameter=1. / grid_intervals, mu_energy_product=mu_bar)
    return fom, mu_bar


def record_results(function, parse, data, mu):
    QoI = function(mu)
    data['num_evals'] += 1
    data['evaluation_points'].append(parse(mu).to_numpy())
    data['evaluations'].append(QoI[0])
    print('.', end='')
    return QoI


def report(result, parse, data, reference_mu):
    if (result.status != 0):
        print('\n failed!')
    else:
        print('\n succeeded!')
        print('  mu_min:    {}'.format(parse(result.x)))
        print('  J(mu_min): {}'.format(result.fun[0]))
        print('  absolute error w.r.t. reference solution: {:.2e}'.format(np.linalg.norm(result.x-reference_mu)))
        print('  num iterations:        {}'.format(result.nit))
        print('  num function calls:    {}'.format(data['num_evals']))
        print('  time:                  {:.5f} seconds'.format(data['time']))
        if 'offline_time' in data:
            print('  offline time:          {:.5f} seconds'.format(data['offline_time']))
    print('')


if __name__ == '__main__':
    run(main)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

from itertools import chain

import numpy as np
import scipy.linalg as spla

from pymor.algorithms.rules import RuleTable, match_generic, match_class_all, match_class_any, match_always
from pymor.core.exceptions import RuleNotMatchingError
from pymor.operators.block import (BlockOperator, BlockRowOperator, BlockColumnOperator, BlockOperatorBase,
                                   BlockDiagonalOperator, SecondOrderModelOperator)
from pymor.operators.constructions import (ZeroOperator, IdentityOperator, VectorArrayOperator, LincombOperator,
                                           LowRankOperator, LowRankUpdatedOperator)
from pymor.vectorarrays.constructions import cat_arrays


def assemble_lincomb(operators, coefficients, solver_options=None, name=None):
    """Try to assemble a linear combination of the given operators.

    Returns a new |Operator| which represents the sum ::

        c_1*O_1 + ... + c_N*O_N

    where `O_i` are |Operators| and `c_i` scalar coefficients.

    This function is called in the :meth:`assemble` method of |LincombOperator| and
    is not intended to be used directly.

    To form the linear combination of backend |Operators| (containing actual matrix data),
    :meth:`~pymor.operators.interface.Operator._assemble_lincomb` will be called
    on the first |Operator| in the linear combination.

    Parameters
    ----------
    operators
        List of |Operators| `O_i` whose linear combination is formed.
    coefficients
        List of the corresponding linear coefficients `c_i`.
    solver_options
        |solver_options| for the assembled operator.
    name
        Name of the assembled operator.

    Returns
    -------
    The assembled |Operator|.
    """
    return AssembleLincombRules(tuple(coefficients), solver_options, name).apply(tuple(operators))


class AssembleLincombRules(RuleTable):
    def __init__(self, coefficients, solver_options, name):
        super().__init__(use_caching=False)
        self.__auto_init(locals())

    @match_always
    def action_zero_coeff(self, ops):
        if all(coeff != 0 for coeff in self.coefficients):
            raise RuleNotMatchingError
        without_zero = [(op, coeff)
                        for op, coeff in zip(ops, self.coefficients)
                        if coeff != 0]
        if len(without_zero) == 0:
            return ZeroOperator(ops[0].range, ops[0].source, name=self.name)
        else:
            new_ops, new_coeffs = zip(*without_zero)
            return assemble_lincomb(new_ops, new_coeffs,
                                    solver_options=self.solver_options, name=self.name)

    @match_class_any(ZeroOperator)
    def action_ZeroOperator(self, ops):
        without_zero = [(op, coeff)
                        for op, coeff in zip(ops, self.coefficients)
                        if not isinstance(op, ZeroOperator)]
        if len(without_zero) == 0:
            return ZeroOperator(ops[0].range, ops[0].source, name=self.name)
        else:
            new_ops, new_coeffs = zip(*without_zero)
            return assemble_lincomb(new_ops, new_coeffs,
                                    solver_options=self.solver_options, name=self.name)

    @match_class_all(IdentityOperator)
    def action_IdentityOperator(self, ops):
        coeff = sum(self.coefficients)
        if coeff == 0:
            return ZeroOperator(ops[0].source, ops[0].source, name=self.name)
        else:
            return LincombOperator([IdentityOperator(ops[0].source, name=self.name)],
                                   [coeff],
                                   name=self.name)

    @match_class_any(BlockOperatorBase)
    @match_class_any(IdentityOperator)
    def action_BlockSpaceIdentityOperator(self, ops):
        new_ops = tuple(
            BlockDiagonalOperator([IdentityOperator(s) for s in op.source.subspaces])
            if isinstance(op, IdentityOperator) else op
            for op in ops if not isinstance(op, ZeroOperator)
        )
        return self.apply(new_ops)

    @match_class_all(VectorArrayOperator)
    def action_VectorArrayOperator(self, ops):
        if not all(op.adjoint == ops[0].adjoint for op in ops):
            raise RuleNotMatchingError

        adjoint = ops[0].adjoint
        assert not self.solver_options

        coeffs = np.conj(self.coefficients) if adjoint else self.coefficients

        if coeffs[0] == 1:
            array = ops[0].array.copy()
        else:
            array = ops[0].array * coeffs[0]
        for op, c in zip(ops[1:], coeffs[1:]):
            array.axpy(c, op.array)

        return VectorArrayOperator(array, adjoint=adjoint, space_id=ops[0].space_id, name=self.name)

    @match_class_any(SecondOrderModelOperator)
    def action_SecondOrderModelOperator(self, ops):
        def is_scaled_iden_like(op):
            if isinstance(op, (ZeroOperator, IdentityOperator)):
                return True
            if (isinstance(op, LincombOperator)
                    and len(op.operators) == 1
                    and isinstance(op.operators[0], IdentityOperator)):
                return True
            return False

        def is_so_op_like(op):
            if isinstance(op, SecondOrderModelOperator):
                return True
            if is_scaled_iden_like(op.blocks[0, 0]) and is_scaled_iden_like(op.blocks[0, 1]):
                return True
            return False

        if not all(is_so_op_like(op) for op in ops):
            raise RuleNotMatchingError

        def so_op_parts(op):
            if isinstance(op, SecondOrderModelOperator):
                return op.alpha, op.beta, op.A, op.B

            def scaled_iden_coeff(op):
                if isinstance(op, ZeroOperator):
                    return 0
                if isinstance(op, IdentityOperator):
                    return 1
                return op.coefficients[0]

            return (scaled_iden_coeff(op.blocks[0, 0]),
                    scaled_iden_coeff(op.blocks[0, 1]),
                    op.blocks[1, 1],
                    op.blocks[1, 0])

        alphas, betas, As, Bs = zip(*(so_op_parts(op) for op in ops))
        alpha = sum(a * b for a, b in zip(self.coefficients, alphas))
        beta = sum(a * b for a, b in zip(self.coefficients, betas))
        A = assemble_lincomb(As, self.coefficients)
        B = assemble_lincomb(Bs, self.coefficients)
        so_op = SecondOrderModelOperator(alpha, beta, A, B)
        return so_op

    @match_class_all(BlockDiagonalOperator)
    def action_BlockDiagonalOperator(self, ops):
        coefficients = self.coefficients
        num_source_blocks = ops[0].num_source_blocks
        blocks = np.empty((num_source_blocks,), dtype=object)
        if len(ops) > 1:
            for i in range(num_source_blocks):
                operators_i = [op.blocks[i, i] for op in ops]
                blocks[i] = assemble_lincomb(operators_i, coefficients,
                                             solver_options=self.solver_options, name=self.name)
                if blocks[i] is None:
                    return None
            return BlockDiagonalOperator(blocks)
        else:
            c = coefficients[0]
            if c == 1:
                return ops[0]
            for i in range(num_source_blocks):
                blocks[i] = ops[0].blocks[i, i] * c
            return BlockDiagonalOperator(blocks)

    @match_class_all(BlockOperatorBase)
    def action_BlockOperatorBase(self, ops):
        coefficients = self.coefficients
        shape = ops[0].blocks.shape
        blocks = np.empty(shape, dtype=object)
        operator_type = ((BlockOperator if ops[0].blocked_source else BlockColumnOperator) if ops[0].blocked_range
                         else BlockRowOperator)
        if len(ops) > 1:
            for (i, j) in np.ndindex(shape):
                operators_ij = [op.blocks[i, j] for op in ops]
                blocks[i, j] = assemble_lincomb(operators_ij, coefficients,
                                                solver_options=self.solver_options, name=self.name)
                if blocks[i, j] is None:
                    return None
            return operator_type(blocks)
        else:
            c = coefficients[0]
            if c == 1:
                return ops[0]
            for (i, j) in np.ndindex(shape):
                blocks[i, j] = ops[0].blocks[i, j] * c
            return operator_type(blocks)

    @match_generic(lambda ops: sum(1 for op in ops if isinstance(op, LowRankOperator)) >= 2)
    def action_merge_low_rank_operators(self, ops):
        low_rank = []
        not_low_rank = []
        for op, coeff in zip(ops, self.coefficients):
            if isinstance(op, LowRankOperator):
                low_rank.append((op, coeff))
            else:
                not_low_rank.append((op, coeff))
        inverted = [op.inverted for op, _ in low_rank]
        if len(inverted) >= 2 and any(inverted) and any(not _ for _ in inverted):
            return None
        inverted = inverted[0]
        left = cat_arrays([op.left for op, _ in low_rank])
        right = cat_arrays([op.right for op, _ in low_rank])
        core = []
        for op, coeff in low_rank:
            core.append(op.core)
            if inverted:
                core[-1] /= coeff
            else:
                core[-1] *= coeff
        core = spla.block_diag(*core)
        new_low_rank_op = LowRankOperator(left, core, right, inverted=inverted)
        if len(not_low_rank) == 0:
            return new_low_rank_op
        else:
            new_ops, new_coeffs = zip(*not_low_rank)
            return assemble_lincomb(chain(new_ops, [new_low_rank_op]), chain(new_coeffs, [1]),
                                    solver_options=self.solver_options, name=self.name)

    @match_generic(lambda ops: len(ops) >= 2)
    @match_class_any(LowRankOperator, LowRankUpdatedOperator)
    def action_merge_into_low_rank_updated_operator(self, ops):
        new_ops = []
        new_lr_ops = []
        new_coeffs = []
        new_lr_coeffs = []
        for op, coeff in zip(ops, self.coefficients):
            if isinstance(op, LowRankOperator):
                new_lr_ops.append(op)
                new_lr_coeffs.append(coeff)
            elif isinstance(op, LowRankUpdatedOperator):
                new_ops.append(op.operators[0])
                new_coeffs.append(coeff * op.coefficients[0])
                new_lr_ops.append(op.operators[1])
                new_lr_coeffs.append(coeff * op.coefficients[1])
            else:
                new_ops.append(op)
                new_coeffs.append(coeff)
        lru_op = assemble_lincomb(new_ops, new_coeffs)
        lru_lr_op = assemble_lincomb(new_lr_ops, new_lr_coeffs)
        lru_lr_coeff = 1
        if isinstance(lru_lr_op, LincombOperator):
            lru_lr_op, lru_lr_coeff = lru_lr_op.operators[0], lru_lr_op.coefficients[0]
        return LowRankUpdatedOperator(lru_op, lru_lr_op, 1, lru_lr_coeff,
                                      solver_options=self.solver_options, name=self.name)

    @match_always
    def action_call_assemble_lincomb_method(self, ops):
        id_coeffs, ops_without_id, coeffs_without_id = [], [], []
        for op, coeff in zip(ops, self.coefficients):
            if isinstance(op, IdentityOperator):
                id_coeffs.append(coeff)
            else:
                ops_without_id.append(op)
                coeffs_without_id.append(coeff)
        id_coeff = sum(id_coeffs)

        op = ops_without_id[0]._assemble_lincomb(ops_without_id, coeffs_without_id, identity_shift=id_coeff,
                                                 solver_options=self.solver_options, name=self.name)

        if not op:
            raise RuleNotMatchingError
        return op

    @match_always
    def action_return_lincomb(self, ops):
        return LincombOperator(ops, self.coefficients, name=self.name)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np
import scipy.linalg as spla

from pymor.algorithms.genericsolvers import _parse_options
from pymor.algorithms.gram_schmidt import gram_schmidt
from pymor.algorithms.lyapunov import _solve_lyap_lrcf_check_args
from pymor.core.defaults import defaults
from pymor.core.logger import getLogger
from pymor.operators.constructions import IdentityOperator
from pymor.tools.random import get_random_state
from pymor.vectorarrays.constructions import cat_arrays


@defaults('lradi_tol', 'lradi_maxiter', 'lradi_shifts', 'projection_shifts_init_maxiter',
          'projection_shifts_init_seed', 'projection_shifts_subspace_columns')
def lyap_lrcf_solver_options(lradi_tol=1e-10,
                             lradi_maxiter=500,
                             lradi_shifts='projection_shifts',
                             projection_shifts_init_maxiter=20,
                             projection_shifts_init_seed=None,
                             projection_shifts_subspace_columns=6):
    """Return available Lyapunov solvers with default options.

    Parameters
    ----------
    lradi_tol
        See :func:`solve_lyap_lrcf`.
    lradi_maxiter
        See :func:`solve_lyap_lrcf`.
    lradi_shifts
        See :func:`solve_lyap_lrcf`.
    projection_shifts_init_maxiter
        See :func:`projection_shifts_init`.
    projection_shifts_init_seed
        See :func:`projection_shifts_init`.
    projection_shifts_subspace_columns
        See :func:`projection_shifts`.

    Returns
    -------
    A dict of available solvers with default solver options.
    """
    return {'lradi': {'type': 'lradi',
                      'tol': lradi_tol,
                      'maxiter': lradi_maxiter,
                      'shifts': lradi_shifts,
                      'shift_options':
                      {'projection_shifts': {'type': 'projection_shifts',
                                             'init_maxiter': projection_shifts_init_maxiter,
                                             'init_seed': projection_shifts_init_seed,
                                             'subspace_columns': projection_shifts_subspace_columns}}}}


def solve_lyap_lrcf(A, E, B, trans=False, options=None):
    """Compute an approximate low-rank solution of a Lyapunov equation.

    See :func:`pymor.algorithms.lyapunov.solve_lyap_lrcf` for a
    general description.

    This function uses the low-rank ADI iteration as described in
    Algorithm 4.3 in :cite:`PK16`.

    Parameters
    ----------
    A
        The non-parametric |Operator| A.
    E
        The non-parametric |Operator| E or `None`.
    B
        The operator B as a |VectorArray| from `A.source`.
    trans
        Whether the first |Operator| in the Lyapunov equation is
        transposed.
    options
        The solver options to use (see
        :func:`lyap_lrcf_solver_options`).

    Returns
    -------
    Z
        Low-rank Cholesky factor of the Lyapunov equation solution,
        |VectorArray| from `A.source`.
    """
    _solve_lyap_lrcf_check_args(A, E, B, trans)
    options = _parse_options(options, lyap_lrcf_solver_options(), 'lradi', None, False)
    logger = getLogger('pymor.algorithms.lradi.solve_lyap_lrcf')

    shift_options = options['shift_options'][options['shifts']]
    if shift_options['type'] == 'projection_shifts':
        init_shifts = projection_shifts_init
        iteration_shifts = projection_shifts
    else:
        raise ValueError('Unknown low-rank ADI shift strategy.')

    if E is None:
        E = IdentityOperator(A.source)

    Z = A.source.empty()
    W = B.copy()

    j = 0
    j_shift = 0
    shifts = init_shifts(A, E, W, shift_options)
    res = np.linalg.norm(W.gramian(), ord=2)
    init_res = res
    Btol = res * options['tol']

    while res > Btol and j < options['maxiter']:
        if shifts[j_shift].imag == 0:
            AaE = A + shifts[j_shift].real * E
            if not trans:
                V = AaE.apply_inverse(W)
                W -= E.apply(V) * (2 * shifts[j_shift].real)
            else:
                V = AaE.apply_inverse_adjoint(W)
                W -= E.apply_adjoint(V) * (2 * shifts[j_shift].real)
            Z.append(V * np.sqrt(-2 * shifts[j_shift].real))
            j += 1
        else:
            AaE = A + shifts[j_shift] * E
            gs = -4 * shifts[j_shift].real
            d = shifts[j_shift].real / shifts[j_shift].imag
            if not trans:
                V = AaE.apply_inverse(W)
                W += E.apply(V.real + V.imag * d) * gs
            else:
                V = AaE.apply_inverse_adjoint(W).conj()
                W += E.apply_adjoint(V.real + V.imag * d) * gs
            g = np.sqrt(gs)
            Z.append((V.real + V.imag * d) * g)
            Z.append(V.imag * (g * np.sqrt(d**2 + 1)))
            j += 2
        j_shift += 1
        res = np.linalg.norm(W.gramian(), ord=2)
        logger.info(f'Relative residual at step {j}: {res/init_res:.5e}')
        if j_shift >= shifts.size:
            shifts = iteration_shifts(A, E, V, Z, shifts, shift_options)
            j_shift = 0

    if res > Btol:
        logger.warning(f'Prescribed relative residual tolerance was not achieved '
                       f'({res/init_res:e} > {options["tol"]:e}) after ' f'{options["maxiter"]} ADI steps.')

    return Z


def projection_shifts_init(A, E, B, shift_options):
    """Find starting projection shifts.

    Uses Galerkin projection on the space spanned by the right-hand side if
    it produces stable shifts.
    Otherwise, uses a randomly generated subspace.
    See :cite:`PK16`, pp. 92-95.

    Parameters
    ----------
    A
        The |Operator| A from the corresponding Lyapunov equation.
    E
        The |Operator| E from the corresponding Lyapunov equation.
    B
        The |VectorArray| B from the corresponding Lyapunov equation.
    shift_options
        The shift options to use (see :func:`lyap_lrcf_solver_options`).

    Returns
    -------
    shifts
        A |NumPy array| containing a set of stable shift parameters.
    """
    random_state = get_random_state(seed=shift_options['init_seed'])
    for i in range(shift_options['init_maxiter']):
        Q = gram_schmidt(B, atol=0, rtol=0)
        shifts = spla.eigvals(A.apply2(Q, Q), E.apply2(Q, Q))
        shifts = shifts[shifts.real < 0]
        if shifts.size == 0:
            # use random subspace instead of span{B} (with same dimensions)
            B = B.random(len(B), distribution='normal', random_state=random_state)
        else:
            return shifts
    raise RuntimeError('Could not generate initial shifts for low-rank ADI iteration.')


def projection_shifts(A, E, V, Z, prev_shifts, shift_options):
    """Find further projection shifts.

    Uses Galerkin projection on spaces spanned by LR-ADI iterates.
    See :cite:`PK16`, pp. 92-95.

    Parameters
    ----------
    A
        The |Operator| A from the corresponding Lyapunov equation.
    E
        The |Operator| E from the corresponding Lyapunov equation.
    V
        A |VectorArray| representing the currently computed iterate.
    Z
        A |VectorArray| representing the current approximate solution.
    prev_shifts
        A |NumPy array| containing the set of all previously used shift
        parameters.
    shift_options
        The shift options to use (see :func:`lyap_lrcf_solver_options`).

    Returns
    -------
    shifts
        A |NumPy array| containing a set of stable shift parameters.
    """
    if shift_options['subspace_columns'] == 1:
        if prev_shifts[-1].imag != 0:
            Q = gram_schmidt(cat_arrays([V.real, V.imag]), atol=0, rtol=0)
        else:
            Q = gram_schmidt(V, atol=0, rtol=0)
    else:
        num_columns = shift_options['subspace_columns'] * len(V)
        Q = gram_schmidt(Z[-num_columns:], atol=0, rtol=0)

    shifts = spla.eigvals(A.apply2(Q, Q), E.apply2(Q, Q))
    shifts = shifts[shifts.real < 0]
    shifts = shifts[shifts.imag >= 0]
    if shifts.size == 0:
        return prev_shifts
    else:
        shifts.imag[-shifts.imag / shifts.real < 1e-12] = 0
        shifts = shifts[np.abs(shifts).argsort()]
        return shifts



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import os
import tempfile

import numpy as np
import pytest
import scipy.sparse as sps

from pymor.models.iosys import LTIModel, SecondOrderModel


def _build_matrices_lti(with_D, with_E):
    A = sps.csc_matrix([[1, 2], [3, 4]])
    B = np.array([[1], [2]])
    C = np.array([[1, 2]])
    D = np.array([[1]]) if with_D else None
    E = np.array([[5, 6], [7, 8]]) if with_E else None
    return A, B, C, D, E


def _test_matrices_lti(A, B, C, D, E,
                       A2, B2, C2, D2, E2,
                       with_D, with_E):
    assert np.allclose(A.toarray(), A2.toarray())
    assert np.allclose(B, B2)
    assert np.allclose(C, C2)
    if with_D:
        assert np.allclose(D, D2)
    else:
        assert D2 is None
    if with_E:
        assert np.allclose(E, E2)
    else:
        assert E2 is None


def _build_matrices_so(with_Cv, with_D):
    M = sps.csc_matrix([[1, 2], [3, 4]])
    E = np.array([[5, 6], [7, 8]])
    K = np.array([[9, 10], [11, 12]])
    B = np.array([[1], [2]])
    Cp = np.array([[1, 2]])
    Cv = np.array([[3, 4]]) if with_Cv else None
    D = np.array([[1]]) if with_D else None
    return M, E, K, B, Cp, Cv, D


def _test_matrices_so(M, E, K, B, Cp, Cv, D,
                      M2, E2, K2, B2, Cp2, Cv2, D2,
                      with_Cv, with_D):
    assert np.allclose(M.toarray(), M2.toarray())
    assert np.allclose(E, E2)
    assert np.allclose(K, K2)
    assert np.allclose(B, B2)
    assert np.allclose(Cp, Cp2)
    if with_Cv:
        assert np.allclose(Cv, Cv2)
    else:
        assert Cv2 is None
    if with_D:
        assert np.allclose(D, D2)
    else:
        assert D2 is None


@pytest.mark.parametrize('with_D', [False, True])
@pytest.mark.parametrize('with_E', [False, True])
def test_matrices_lti(with_D, with_E):
    matrices = _build_matrices_lti(with_D, with_E)

    lti = LTIModel.from_matrices(*matrices)
    matrices2 = lti.to_matrices()

    _test_matrices_lti(*matrices, *matrices2, with_D, with_E)


@pytest.mark.parametrize('with_D', [False, True])
@pytest.mark.parametrize('with_E', [False, True])
def test_files_lti(with_D, with_E):
    matrices = _build_matrices_lti(with_D, with_E)

    lti = LTIModel.from_matrices(*matrices)
    with tempfile.TemporaryDirectory() as tmpdirname:
        files = (
            os.path.join(tmpdirname, 'A.mtx'),
            os.path.join(tmpdirname, 'B.mtx'),
            os.path.join(tmpdirname, 'C.mtx'),
            os.path.join(tmpdirname, 'D.mtx') if with_D else None,
            os.path.join(tmpdirname, 'E.mtx') if with_E else None,
        )
        lti.to_files(*files)
        lti2 = LTIModel.from_files(*files)
    matrices2 = lti2.to_matrices()

    _test_matrices_lti(*matrices, *matrices2, with_D, with_E)


@pytest.mark.parametrize('with_D', [False, True])
@pytest.mark.parametrize('with_E', [False, True])
def test_mat_file_lti(with_D, with_E):
    matrices = _build_matrices_lti(with_D, with_E)

    lti = LTIModel.from_matrices(*matrices)
    with tempfile.TemporaryDirectory() as tmpdirname:
        file_name = os.path.join(tmpdirname, 'lti')
        lti.to_mat_file(file_name)
        lti2 = LTIModel.from_mat_file(file_name)
    matrices2 = lti2.to_matrices()

    _test_matrices_lti(*matrices, *matrices2, with_D, with_E)


@pytest.mark.parametrize('with_D', [False, True])
@pytest.mark.parametrize('with_E', [False, True])
def test_abcde_files(with_D, with_E):
    matrices = _build_matrices_lti(with_D, with_E)

    lti = LTIModel.from_matrices(*matrices)
    with tempfile.TemporaryDirectory() as tmpdirname:
        files_basename = os.path.join(tmpdirname, 'lti')
        lti.to_abcde_files(files_basename)
        lti2 = LTIModel.from_abcde_files(files_basename)
    matrices2 = lti2.to_matrices()

    _test_matrices_lti(*matrices, *matrices2, with_D, with_E)


@pytest.mark.parametrize('with_Cv', [False, True])
@pytest.mark.parametrize('with_D', [False, True])
def test_matrices_so(with_Cv, with_D):
    matrices = _build_matrices_so(with_Cv, with_D)

    som = SecondOrderModel.from_matrices(*matrices)
    matrices2 = som.to_matrices()

    _test_matrices_so(*matrices, *matrices2, with_Cv, with_D)


@pytest.mark.parametrize('with_Cv', [False, True])
@pytest.mark.parametrize('with_D', [False, True])
def test_files_so(with_Cv, with_D):
    matrices = _build_matrices_so(with_Cv, with_D)

    som = SecondOrderModel.from_matrices(*matrices)
    with tempfile.TemporaryDirectory() as tmpdirname:
        files = (
            os.path.join(tmpdirname, 'M.mtx'),
            os.path.join(tmpdirname, 'E.mtx'),
            os.path.join(tmpdirname, 'K.mtx'),
            os.path.join(tmpdirname, 'B.mtx'),
            os.path.join(tmpdirname, 'Cp.mtx'),
            os.path.join(tmpdirname, 'Cv.mtx') if with_Cv else None,
            os.path.join(tmpdirname, 'D.mtx') if with_D else None,
        )
        som.to_files(*files)
        som2 = SecondOrderModel.from_files(*files)
    matrices2 = som2.to_matrices()

    _test_matrices_so(*matrices, *matrices2, with_Cv, with_D)



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import numpy as np
import scipy.linalg as spla
import scipy.sparse as sps
import scipy.sparse.linalg as spsla

from pymor.algorithms.rules import RuleTable, match_class
from pymor.operators.block import BlockOperatorBase
from pymor.operators.constructions import (AdjointOperator, ComponentProjectionOperator, ConcatenationOperator,
                                           IdentityOperator, LincombOperator, LowRankOperator, LowRankUpdatedOperator,
                                           VectorArrayOperator, ZeroOperator)
from pymor.operators.numpy import NumpyMatrixOperator


def to_matrix(op, format=None, mu=None):
    """Convert a linear |Operator| to a matrix.

    Parameters
    ----------
    op
        The |Operator| to convert.
    format
        Format of the resulting matrix: |NumPy array| if 'dense',
        otherwise the appropriate |SciPy spmatrix|.
        If `None`, a choice between dense and sparse format is
        automatically made.
    mu
        The |parameter values| for which to convert `op`.

    Returns
    -------
    res
        The matrix equivalent to `op`.
    """
    assert format is None or format in ('dense', 'bsr', 'coo', 'csc', 'csr', 'dia', 'dok', 'lil')
    op = op.assemble(mu)
    return ToMatrixRules(format, mu).apply(op)


class ToMatrixRules(RuleTable):

    def __init__(self, format, mu):
        super().__init__()
        self.__auto_init(locals())

    @match_class(NumpyMatrixOperator)
    def action_NumpyMatrixOperator(self, op):
        format = self.format
        if format is None:
            return op.matrix
        elif format == 'dense':
            if not op.sparse:
                return op.matrix
            else:
                return op.matrix.toarray()
        else:
            if not op.sparse:
                return getattr(sps, format + '_matrix')(op.matrix)
            else:
                return op.matrix.asformat(format)

    @match_class(BlockOperatorBase)
    def action_BlockOperator(self, op):
        format = self.format
        op_blocks = op.blocks
        mat_blocks = [[] for i in range(op.num_range_blocks)]
        is_dense = True
        for i in range(op.num_range_blocks):
            for j in range(op.num_source_blocks):
                mat_ij = self.apply(op_blocks[i, j])
                if sps.issparse(mat_ij):
                    is_dense = False
                mat_blocks[i].append(mat_ij)
        if format is None and is_dense or format == 'dense':
            return np.asarray(np.bmat(mat_blocks))
        else:
            return sps.bmat(mat_blocks, format=format)

    @match_class(AdjointOperator)
    def action_AdjointOperator(self, op):
        format = self.format
        res = self.apply(op.operator).T.conj()
        if op.range_product is not None:
            range_product = self.apply(op.range_product)
            if format is None and not sps.issparse(res) and sps.issparse(range_product):
                res = range_product.T.dot(res.T).T
            else:
                res = res.dot(range_product)
        if op.source_product is not None:
            source_product = self.apply(op.source_product)
            if not sps.issparse(source_product):
                res = spla.solve(source_product, res)
            else:
                res = spsla.spsolve(source_product, res)
        if format is not None and format != 'dense':
            res = getattr(sps, format + '_matrix')(res)
        return res

    @match_class(ComponentProjectionOperator)
    def action_ComponentProjectionOperator(self, op):
        format = self.format
        if format == 'dense':
            res = np.zeros((op.range.dim, op.source.dim))
            for i, j in enumerate(op.components):
                res[i, j] = 1
        else:
            data = np.ones((op.range.dim,))
            i = np.arange(op.range.dim)
            j = op.components
            res = sps.coo_matrix((data, (i, j)), shape=(op.range.dim, op.source.dim))
            res = res.asformat(format if format else 'csc')
        return res

    @match_class(ConcatenationOperator)
    def action_ConcatenationOperator(self, op):
        mats = [self.apply(o) for o in op.operators]
        while len(mats) > 1:
            if self.format is None and not sps.issparse(mats[-2]) and sps.issparse(mats[-1]):
                mats = mats[:-2] + [mats[-1].T.dot(mats[-2].T).T]
            else:
                mats = mats[:-2] + [mats[-2].dot(mats[-1])]
        return mats[0]

    @match_class(IdentityOperator)
    def action_IdentityOperator(self, op):
        format = self.format
        if format == 'dense':
            return np.eye(op.source.dim)
        else:
            return sps.eye(op.source.dim, format=format if format else 'csc')

    @match_class(LincombOperator)
    def action_LincombOperator(self, op):
        op_coefficients = op.evaluate_coefficients(self.mu)
        res = op_coefficients[0] * self.apply(op.operators[0])
        for i in range(1, len(op.operators)):
            res = res + op_coefficients[i] * self.apply(op.operators[i])
        return res

    @match_class(LowRankOperator)
    def action_LowRankOperator(self, op):
        format = self.format
        if not op.inverted:
            res = op.left.to_numpy().T @ op.core @ op.right.to_numpy()
        else:
            res = op.left.to_numpy().T @ spla.solve(op.core, op.right.to_numpy())
        if format is not None and format != 'dense':
            res = getattr(sps, format + '_matrix')(res)
        return res

    @match_class(LowRankUpdatedOperator)
    def action_LowRankUpdatedOperator(self, op):
        return op.coeff * self.apply(op.operator) + op.lr_coeff * self.apply(op.lr_operator)

    @match_class(VectorArrayOperator)
    def action_VectorArrayOperator(self, op):
        format = self.format
        res = op.array.conj().to_numpy() if op.adjoint else op.array.to_numpy().T
        if format is not None and format != 'dense':
            res = getattr(sps, format + '_matrix')(res)
        return res

    @match_class(ZeroOperator)
    def action_ZeroOperator(self, op):
        format = self.format
        if format is None:
            return sps.csc_matrix((op.range.dim, op.source.dim))
        elif format == 'dense':
            return np.zeros((op.range.dim, op.source.dim))
        else:
            return getattr(sps, format + '_matrix')((op.range.dim, op.source.dim))



# This file is part of the pyMOR project (https://www.pymor.org).
# Copyright 2013-2021 pyMOR developers and contributors. All rights reserved.
# License: BSD 2-Clause License (https://opensource.org/licenses/BSD-2-Clause)

import operator
import warnings

import numpy as np

from pymor.core.defaults import defaults


@defaults('rtol', 'atol')
def float_cmp(x, y, rtol=1e-14, atol=1e-14):
    """Compare x and y component-wise for almost equality.

    For scalars we define almost equality as ::

       float_cmp(x,y) <=> |x - y| <= atol + |y|*rtol

    .. note::
       Numpy's :meth:`~numpy.allclose` method uses the same definition but
       treats arrays containing infinities as close if the infinities are
       at the same places and all other entries are close.
       In our definition, arrays containing infinities can never be close
       which seems more appropriate in most cases.

    Parameters
    ----------
    x, y
        |NumPy arrays| to be compared. Have to be broadcastable to the same shape.
    rtol
        The relative tolerance.
    atol
        The absolute tolerance.
    """
    return np.abs(x - y) <= atol + np.abs(y) * rtol


def float_cmp_all(x, y, rtol=None, atol=None):
    """Compare x and y for almost equality.

    Returns `True` if all components of `x` are almost equal to the corresponding
    components of `y`.

    See :meth:`float_cmp`.
    """
    return np.all(float_cmp(x, y, rtol, atol))


def bounded(lower, upper, x, rtol=None, atol=None):
    """Check if x is strictly in bounds (lower, upper) or float_compares equal to lower or upper.

    Parameters
    ----------
    lower
        Lower bound
    upper
        Upper bound
    x
        value to check
    rtol
        relative tolerance for float_cmp
    atol
        absolute tolerance for float_cmp
    """
    return (lower < x < upper) or float_cmp(x, lower, rtol, atol) or float_cmp(x, upper, rtol, atol)


@defaults('rtol', 'atol')
def almost_less(x, y, rtol=1e-14, atol=1e-14):
    """Component-wise check if x <= y up to a given tolerance.

    For scalars the check is given by  ::

       almost_less(x, y) <=> (x - y  <= atol + |y| * rtol)

    Parameters
    ----------
    x, y
        |NumPy arrays| to be compared. Have to be broadcastable to the same shape.
    rtol
        The relative tolerance.
    atol
        The absolute tolerance.
    """
    return x - y <= atol + np.abs(y) * rtol
